import json
import uuid
from math import log, isnan

from dash import Dash, html, dcc, Input, Output
import pandas as pd
import plotly.express as px

from cxdash import CXDashElement

_g_external_stylesheets = ["https://codepen.io/chriddyp/pen/bWLwgP.css"]

_g_app = Dash(__name__, external_stylesheets=_g_external_stylesheets)

_g_country_indicators_df = pd.read_csv(
    "https://plotly.github.io/datasets/country_indicators.csv"
)


_g_app.layout = html.Div(
    [
        html.H1(
            "Mixed CanvasXpress and Plotly Graph Example",
            style={"width": "100%", "text-align": "center"},
        ),
        html.Hr(),
        html.Div(
            """
            This Dash application demonstrates a mixture of CanvasXpress and Plotly elements coordinating data
            exchange and UI refresh.  The CXDashElement class is used for preparing CanvasXpress charts, which is
            a straightforward mechanism to use but can require additional work on the developer's part to process 
            or format data, configurations, and events.  Also see the advanced CanvasXpress edition of this example.
            """
        ),
        html.Hr(),
        html.Div(
            [
                html.Div(style={"display": "inline-block", "width": "5%"}),
                html.Div(
                    [
                        html.Div(
                            [
                                dcc.Dropdown(
                                    _g_country_indicators_df["Indicator Name"].unique(),
                                    "Life expectancy at birth, total (years)",
                                    id="crossfilter-yaxis-column",
                                    style={"display": "inline-block", "width": "85%"},
                                ),
                                dcc.RadioItems(
                                    ["Linear", "Log"],
                                    "Linear",
                                    id="crossfilter-yaxis-type",
                                    style={"display": "inline-block", "width": "15%"},
                                ),
                            ],
                            style={"display": "inline-block", "width": "100%"},
                        ),
                    ],
                    style={"display": "inline-block", "width": "40%"},
                ),
                html.Div(style={"display": "inline-block", "width": "10%"}),
                html.Div(
                    [
                        html.Div(
                            [
                                dcc.Dropdown(
                                    _g_country_indicators_df["Indicator Name"].unique(),
                                    "Fertility rate, total (births per woman)",
                                    id="crossfilter-xaxis-column",
                                    style={"display": "inline-block", "width": "85%"},
                                ),
                                dcc.RadioItems(
                                    ["Linear", "Log"],
                                    "Linear",
                                    id="crossfilter-xaxis-type",
                                    style={"display": "inline-block", "width": "15%"},
                                ),
                            ],
                            style={"display": "inline-block", "width": "100%"},
                        ),
                    ],
                    style={"display": "inline-block", "width": "40%"},
                ),
                html.Div(style={"display": "inline-block", "width": "5%"}),
            ],
        ),
        html.Div(
            [
                html.Div(style={"display": "inline-block", "width": "2%"}),
                html.Div(
                    dcc.Slider(
                        _g_country_indicators_df["Year"].min(),
                        _g_country_indicators_df["Year"].max(),
                        step=None,
                        id="crossfilter-year--slider",
                        value=_g_country_indicators_df["Year"].max(),
                        marks={
                            str(year): str(year)
                            for year in _g_country_indicators_df["Year"].unique()
                        },
                    ),
                    style={"display": "inline-block", "width": "93%"},
                ),
                html.Div(style={"display": "inline-block", "width": "2%"}),
            ],
        ),
        html.Hr(),
        html.Div(
            [
                html.Div(
                    [
                        dcc.Graph(
                            id="crossfilter-indicator-scatter",
                            hoverData={"points": [{"customdata": "Japan"}]},
                        ),
                    ],
                    style={"display": "inline-block", "width": "50%"},
                ),
                html.Div(
                    [
                        html.Div(
                            [
                                html.Div(
                                    id="y-time-series",
                                    style={"display": "inline-block", "width": "100%"},
                                ),
                            ],
                        ),
                        html.Div(
                            [
                                html.Div(
                                    id="x-time-series",
                                    style={"display": "inline-block", "width": "100%"},
                                ),
                            ],
                        ),
                    ],
                    style={"display": "inline-block", "width": "50%"},
                ),
            ],
        ),
        html.Hr(),
    ]
)


@_g_app.callback(
    Output("crossfilter-indicator-scatter", "figure"),
    Input("crossfilter-xaxis-column", "value"),
    Input("crossfilter-yaxis-column", "value"),
    Input("crossfilter-xaxis-type", "value"),
    Input("crossfilter-yaxis-type", "value"),
    Input("crossfilter-year--slider", "value"),
)
def update_graph(
    xaxis_column_name, yaxis_column_name, xaxis_type, yaxis_type, year_value
):
    dff = _g_country_indicators_df[_g_country_indicators_df["Year"] == year_value]

    fig = px.scatter(
        x=dff[dff["Indicator Name"] == xaxis_column_name]["Value"],
        y=dff[dff["Indicator Name"] == yaxis_column_name]["Value"],
        hover_name=dff[dff["Indicator Name"] == yaxis_column_name]["Country Name"],
    )

    fig.update_traces(
        customdata=dff[dff["Indicator Name"] == yaxis_column_name]["Country Name"]
    )

    fig.update_xaxes(
        title=xaxis_column_name, type="linear" if xaxis_type == "Linear" else "log"
    )

    fig.update_yaxes(
        title=yaxis_column_name, type="linear" if yaxis_type == "Linear" else "log"
    )

    fig.update_layout(margin={"l": 40, "b": 40, "t": 10, "r": 0}, hovermode="closest")

    return fig


def adjust_value(value, type):
    if isnan(value):
        return None
    elif type == "Linear":
        return value
    else:
        return log(value)


def format_data_point(data_point, axis_type):
    return [int(data_point[0]), adjust_value(float(round(data_point[1], 2)), axis_type)]


def create_time_series(id, dff, axis_type, title):

    data = {
        "y": {
            "data": [
                format_data_point(
                    dff[["Year", "Value"]].iloc[i].values.tolist(), axis_type
                )
                for i in range(dff["Value"].count())
            ],
            "smps": ["Date", "Value"],
        },
        "z": {"Series": ["Value" for i in range(dff["Value"].count())]},
    }

    config = {
        "graphOrientation": "vertical",
        "graphType": "Scatter2D",
        "lineBy": "Series",
        "smpLabelRotate": 90,
        "yAxisTitle": " ",
        "xAxisTitle": "Year",
        "xAxis2Show": False,
        "showLegend": False,
        "theme": "CanvasXpress",
        "title": pd.unique(dff["Indicator Name"].values)[0],
        "titleAlign": "left",
        "titleScaleFontFactor": 0.65,
        "subtitle": pd.unique(dff["Country Name"].values)[0],
        "subtitleAlign": "left",
        "subtitleScaleFontFactor": 0.5,
    }

    events = """
    {
        "mousemove": function(o, e, t) {
            t.showInfoSpan(e, '<b>' + o.y.data[0][0] + '</b>: ' + o.y.data[0][1]);
        },
        "mouseout": function(o, e, t) {},
    }
    """

    chart = CXDashElement(
        id=str(uuid.uuid4()),
        data=json.dumps(data),
        config=json.dumps(config),
        events=events,
        height="225",
    )

    return [
        chart,
    ]


@_g_app.callback(
    Output("x-time-series", "children"),
    Input("crossfilter-indicator-scatter", "hoverData"),
    Input("crossfilter-xaxis-column", "value"),
    Input("crossfilter-xaxis-type", "value"),
)
def update_y_timeseries(hoverData, xaxis_column_name, axis_type):
    country_name = hoverData["points"][0]["customdata"]
    dff = _g_country_indicators_df[
        _g_country_indicators_df["Country Name"] == country_name
    ]
    dff = dff[dff["Indicator Name"] == xaxis_column_name]
    title = "{} {}".format(country_name, xaxis_column_name)
    return create_time_series("x", dff, axis_type, title)


@_g_app.callback(
    Output("y-time-series", "children"),
    Input("crossfilter-indicator-scatter", "hoverData"),
    Input("crossfilter-yaxis-column", "value"),
    Input("crossfilter-yaxis-type", "value"),
)
def update_x_timeseries(hoverData, yaxis_column_name, axis_type):
    dff = _g_country_indicators_df[
        _g_country_indicators_df["Country Name"] == hoverData["points"][0]["customdata"]
    ]
    dff = dff[dff["Indicator Name"] == yaxis_column_name]
    return create_time_series("y", dff, axis_type, yaxis_column_name)


if __name__ == "__main__":
    _g_app.run_server(debug=False)



import json
from functools import total_ordering
from typing import Union

from pandas import DataFrame

from canvasxpress.data.base import CXMatrixData


@total_ordering
class CXDataframeData(CXMatrixData):
    """
    A CXData class dedicated to processing Python DataFrame, matrix-structured
     data.
    """

    __data: DataFrame = DataFrame()
    """
    The data managed by an object of this class.
    """

    @property
    def dataframe(self) -> DataFrame:
        """
        Provides the data managed by the object.
        :returns: `DataFrame` The managed data.
        """
        return self.__data

    @dataframe.setter
    def dataframe(self, value: Union[DataFrame, None] = None) -> None:
        """
        Sets the dataframe managed by the object.
        :param value: `Union[DataFrame, None]`
            `None` results in an empty `DataFrame`.  A deepcopy will be made of
            `DataFrame` values.
        """
        if not isinstance(value, (DataFrame, type(None))):
            raise TypeError("The assignment value must be a DataFrame or None.")

        candidate = value
        if candidate is None:
            candidate = DataFrame()

        self.data = candidate

    @property
    def data(self) -> dict:
        """
        Provides the data managed by the object.
        :returns: `DataFrame` The managed data.
        """
        return self.dataframe.to_dict(orient="list")

    @data.setter
    def data(self, value: Union["CXDataframeData", DataFrame, None] = None) -> None:
        """
        Sets the dataframe managed by the object.
        :param value: `Union['CXDataframeData', DataFrame, dict, str, None]`
            `None` results in an empty `DataFrame`.  A deepcopy will be made of
            `DataFrame` or equivalent values.
        """
        if value is None:
            self.__data = DataFrame()

        elif not isinstance(value, (CXDataframeData, DataFrame, type(None))):
            raise TypeError("The assignment value must be a DataFrame or None.")

        elif isinstance(value, CXDataframeData):
            self.__data = value.dataframe.copy(deep=True)

        elif isinstance(value, DataFrame):
            self.__data = value.copy(deep=True)

        else:
            self.__data = DataFrame()

    def get_raw_dict_form(self) -> dict:
        """ "
        Provides a simple dict perspective of the data with no metadata or other
        contextual transformations performed.  For example, if the data is
        natively in `dict` form then it would be passed-through with no
        modification or enhancement.

        This implementation provides matrix data formatted in a `dict` object
        with `DataFrame.to_dict('split')` behaviour.

        :returns: `dict`
            The `dict` perspective of the data with as little modification or
            interpretation as is reasonable.
        """
        return self.__data.to_dict(orient="split")

    def render_to_dict(self, **kwargs) -> dict:
        """
        Provides a dict representation of the data.
        :returns: `dict`
            The data in `dict` form.
        """
        candidate = self.get_raw_dict_form()
        return candidate

    def __init__(self, data: Union["CXDataframeData", DataFrame, None] = None) -> None:
        """
        Initializes the CXData object with data.  Only `DataFrame` or compatible
         data types are accepted.
        :param data: `Union['CXDataframeData', DataFrame, dict, str, None]`
            `None` to initialize with an empty `DataFrame`, or a `DataFrame`
            like object to assign mapped data.
        """
        super().__init__(data)
        self.data = data

    def __copy__(self) -> "CXDataframeData":
        """
        *copy constructor* that returns a copy of the CXDataframeData object.
        :returns: `CXDataframeData`
            A copy of the wrapping object.
        """
        return self.__class__(self.dataframe)

    def __deepcopy__(self, memo) -> "CXDataframeData":
        """
        *deepcopy constructor* that returns a copy of the CXDataframeData object.
        :returns: `CXDataframeData` A copy of the wrapping object and deepcopy of
            the tracked data.
        """
        return self.__class__(self.dataframe)

    def __lt__(self, other: "CXDataframeData") -> bool:
        """
        *less than* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXDataframeData` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXDataframeData` object then False
            <li> If `other` is a `CXDataframeData` object then True of all
                `CXDataframeData` aspects are also less than the data tracked by
                `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXDataframeData):
            return False

        else:
            self_c = self.dataframe.columns.unique()
            other_c = other.dataframe.columns.unique()

            if len(self_c) < len(other_c):
                return True

            elif len(self_c) > len(other_c):
                return False

            else:
                for i in [s for s in self_c if s not in other_c]:
                    if any([i < o for o in other_c]):
                        return True

                return self.dataframe.lt(other.dataframe).all(axis=None)

    def __eq__(self, other: "CXDataframeData") -> bool:
        """
        *equals* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXDataframeData` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXDataframeData` object then False
            <li> If `other` is a `CXDataframeData` object then True of all
                `CXDataframeData` aspects are also less than the data tracked by
                `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXDataframeData):
            return False

        else:
            self_c = self.dataframe.columns.unique()
            other_c = other.dataframe.columns.unique()

            if len(self_c) != len(other_c):
                return False

            if any([s not in other_c for s in self_c]):
                return False

            return self.dataframe.eq(other.dataframe).all(axis=None)

    def __str__(self) -> str:
        """
        *str* function.  Converts the CXDataframeData object into a JSON
        representation.
        :returns" `str` JSON form of the `CXDataframeData`.
        """
        if self.dataframe is None:
            return str(None)

        else:
            return json.dumps(self.render_to_dict())

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXDataframeData object into a pickle
         string that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        candidate = (
            f"CXDataframeData("
            f"data=pandas.read_csv("
            f'StringIO("""{self.dataframe.to_csv(index=True)}"""),'
            f"index_col=0))"
        )
        candidate = candidate.replace("Infinity", "float('inf')")
        candidate = candidate.replace("NaN", "float('nan')")

        return candidate


def merge_dataframes_into_xyz_object(
    data: CXDataframeData,
    sample_annotation: CXDataframeData = None,
    variable_annotation: CXDataframeData = None,
) -> dict:
    """
    Converts a set of DataFrame like objects into an XYZ dict.
    """
    xyz_data = {}

    if data is not None:
        xyz_data["y"] = {}
        xyz_data["y"]["data"] = data.dataframe.values.tolist()
        xyz_data["y"]["smps"] = data.dataframe.columns.tolist()
        xyz_data["y"]["vars"] = data.dataframe.index.tolist()

    if sample_annotation is not None and sample_annotation.dataframe.size > 0:

        xyz_data["x"] = {}

        try:

            # 1.  Scan first column - use values in each row.
            # 2.  Else use first row - use values in each column
            # 3.  Else scan index for a matching sample identifier - use values in each row.
            # 4.  Else scan the header - use values in each column

            # 1.  Scan first column - use values in each row.
            found_strategy = True
            for field in sample_annotation.dataframe[
                sample_annotation.dataframe.columns[0]
            ].values:
                if field not in xyz_data["y"]["smps"]:
                    found_strategy = False
                    break

            if found_strategy:
                for row_index in range(sample_annotation.dataframe.shape[0]):
                    key = sample_annotation.dataframe.iloc[row_index, 0]
                    if not isinstance(key, str):
                        x_key = key.item()
                    xyz_data["x"][key] = sample_annotation.dataframe.iloc[row_index][1:]

            # 2.  Else use first row - use values in each column
            else:
                found_strategy = True
                for column_index in range(sample_annotation.dataframe.shape[1]):
                    if (
                        sample_annotation.dataframe.iloc[0, column_index]
                        not in xyz_data["y"]["smps"]
                    ):
                        found_strategy = False
                        break

                if found_strategy:
                    for column_index in range(sample_annotation.dataframe.shape[1]):
                        column = sample_annotation.dataframe.columns[column_index]
                        meta_name = sample_annotation.dataframe[column][0]
                        if not isinstance(meta_name, str):
                            meta_name = meta_name.item()
                        meta_data = sample_annotation.dataframe[column][1:]
                        xyz_data["x"][meta_name] = meta_data

                # 3.  Else scan header for a matching sample identifier - use values in each column.
                else:
                    found_strategy = True
                    for header in sample_annotation.dataframe.columns.values:
                        if header not in xyz_data["y"]["smps"]:
                            found_strategy = False
                            break

                    if found_strategy:
                        for column_index, column in enumerate(
                            sample_annotation.dataframe.columns.values
                        ):
                            xyz_data["x"][
                                column if isinstance(column, str) else column.item()
                            ] = sample_annotation.dataframe[column]

                    # 4.  Else scan the index - use values in each row
                    else:
                        for index in sample_annotation.dataframe.index.values:
                            xyz_data["x"][
                                index if isinstance(index, str) else index.item()
                            ] = sample_annotation.dataframe.loc[index]

        except Exception as e:
            raise ValueError(
                "Sample Annotation data (x) cannot be parsed or aligned with Chart data (y)"
            )

    if variable_annotation is not None and variable_annotation.dataframe.size > 0:

        xyz_data["z"] = {}

        try:

            # 1.  Scan first column - use values in each row.
            # 2.  Else use first row - use values in each column
            # 3.  Else scan index for a matching sample identifier - use values in each row.
            # 4.  Else scan the header - use values in each column

            # 1.  Scan first column - use values in each row.
            found_strategy = True
            for field in variable_annotation.dataframe[
                variable_annotation.dataframe.columns[0]
            ].values:
                if field not in xyz_data["y"]["vars"]:
                    found_strategy = False
                    break

            if found_strategy:
                for row_index in range(variable_annotation.dataframe.shape[0]):
                    key = variable_annotation.dataframe.iloc[row_index, 0]
                    if not isinstance(key, str):
                        key = key.item()
                    xyz_data["z"][key] = variable_annotation.dataframe.iloc[row_index][
                        1:
                    ]

            # 2.  Else use first row - use values in each column
            else:
                found_strategy = True
                for column_index in range(variable_annotation.dataframe.shape[1]):
                    if (
                        variable_annotation.dataframe.iloc[0, column_index]
                        not in xyz_data["y"]["vars"]
                    ):
                        found_strategy = False
                        break

                if found_strategy:
                    for column_index in range(variable_annotation.dataframe.shape[1]):
                        column = sample_annotation.dataframe.columns[column_index]
                        meta_name = sample_annotation.dataframe[column][0]
                        if not isinstance(meta_name, str):
                            meta_name = meta_name.item()
                        meta_data = sample_annotation.dataframe[column][1:]
                        xyz_data["z"][meta_name] = meta_data

                # 3.  Else scan header for a matching sample identifier - use values in each column.
                else:
                    found_strategy = True
                    for header in variable_annotation.dataframe.columns.values:
                        if header not in xyz_data["y"]["vars"]:
                            found_strategy = False
                            break

                    if found_strategy:
                        for column_index, column in enumerate(
                            variable_annotation.dataframe.columns.values
                        ):
                            xyz_data["z"][
                                column if isinstance(column, str) else column.item()
                            ] = variable_annotation.dataframe[column]

                    # 4.  Else scan the index - use values in each rpw
                    else:
                        for index in variable_annotation.dataframe.index.values:
                            xyz_data["z"][
                                index if isinstance(index, str) else index.item()
                            ] = variable_annotation.dataframe.loc[index]

        except Exception as e:
            raise ValueError(
                "Variable Annotation data (z) cannot be parsed or aligned with Chart data (y)"
            )

    # Convert to Python native types for serialization

    if "x" in xyz_data:
        for annotation in xyz_data["x"].keys():
            xyz_data["x"][annotation] = [
                element if isinstance(element, str) else element.item()
                for element in xyz_data["x"][annotation]
            ]

    if "z" in xyz_data:
        for annotation in xyz_data["z"].keys():
            xyz_data["z"][annotation] = [
                element if isinstance(element, str) else element.item()
                for element in xyz_data["z"][annotation]
            ]

    return xyz_data



import logging
import warnings

from shiny import ui

from canvasxpress.canvas import CanvasXpress

_cx_js_intermixed_template = """
<script type="text/javascript">
    @code@
</script>
"""

_cx_html_intermixed_template = """
<div>
    @canvasxpress_license@
    @canvase@
    @js_functions@
</div>
"""


def output_canvasxpress(id: str) -> ui.TagList:
    """
    Establishes an output reactive placeholder into which a CanvasXpress chart can be rendered.
    """
    if id is None:
        raise ValueError(
            "output_canvasxpress requires that id be of type str and not None."
        )

    elif not isinstance(id, str):
        raise TypeError("output_canvasxpress requires that id be of type str.")

    else:
        css_url = CanvasXpress.css_library_url()
        js_url = CanvasXpress.js_library_url()

        return ui.TagList(
            ui.head_content(
                ui.tags.link(href=css_url, rel="stylesheet"),
                ui.tags.script(src=js_url),
            ),
            ui.output_ui(id),
        )


class CXShinyWidget(object):
    """
    A Shiny for Python compatible class that can be used with Shiny syntax to establish CanvasXpress charts in the
    shiny UI.
    """

    def __init__(self, canvas: CanvasXpress):
        """
        Initializes this object with a valid reference to a CanvasXpress object.  An invalid reference will result in
        a ValueError (`None`) or TypeError (an object not of the type `CanvasXpress`) exception.
        :param canvas: `CanvasXpress` - A valid CanvasXpress object to be rendered in the shiny UI.
        """
        if canvas is None:
            raise ValueError("canvas must be an instance of CanvasXpress")

        elif not isinstance(canvas, CanvasXpress):
            raise TypeError("canvas must be an instance of CanvasXpress")

        else:
            self._canvas = canvas

    def _repr_html_(self):
        """
        Renders the object as Shiny compliant HTML.
        """
        # Get the HTML and JS assets.
        html_parts: dict = self._canvas.render_to_html_parts()

        # Provide the taglist.
        components = ui.TagList(
            [
                ui.div(
                    ui.HTML(
                        _cx_html_intermixed_template.replace(
                            "@canvasxpress_license@",
                            html_parts.get("cx_license", ""),
                        )
                        .replace(
                            "@canvase@",
                            html_parts["cx_canvas"],
                        )
                        .replace(
                            "@js_functions@",
                            _cx_js_intermixed_template.replace(
                                "@code@",
                                html_parts["cx_js"],
                            ),
                        )
                    ),
                ),
            ]
        )

        # Generate the chart DIV and provide it for rendering.
        return components.get_html_string()

    def _repr_rstudio_viewer_(self):
        """
        Renders the object as Shiny compliant HTML.
        """
        try:
            warnings.filterwarnings("ignore")
            logging.getLogger("rpy2.rinterface_lib.embedded").setLevel(logging.ERROR)

            from rpy2 import robjects
        except:
            robjects = None

        # Get the header assets.
        css_url = CanvasXpress.css_library_url()
        js_url = CanvasXpress.js_library_url()

        # Get the HTML and JS assets.
        html_parts: dict = self._canvas.render_to_html_parts()

        # Provide the taglist.
        components = [
            ui.div(
                ui.head_content(
                    ui.tags.link(href=css_url, rel="stylesheet"),
                    ui.tags.script(src=js_url),
                ),
                ui.HTML(
                    _cx_html_intermixed_template.replace(
                        "@canvasxpress_license@",
                        html_parts.get("cx_license", ""),
                    )
                    .replace(
                        "@canvase@",
                        html_parts["cx_canvas"],
                    )
                    .replace(
                        "@js_functions@",
                        _cx_js_intermixed_template.replace(
                            "@code@",
                            html_parts["cx_js"],
                        ),
                    )
                ),
            ),
        ]

        # Generate the chart DIV and provide it for rendering.
        ui_components = ui.TagList(components)
        html = ui_components.get_html_string().replace("\n", "")

        robjects.globalenv["html"] = html
        robjects.r(
            """
            tf <- tempfile(fileext = ".html")
            writeLines(html, tf)
            rstudioapi::viewer(tf)
            """
        )



from math import log, isnan

from dash import Dash, html, dcc, Input, Output
import pandas as pd
import plotly.express as px

from canvasxpress.canvas import CanvasXpress
from canvasxpress.data.matrix import CXDataframeData
from canvasxpress.js.function import CXEvent
from canvasxpress.render.dash import CXElementFactory

_g_external_stylesheets = ["https://codepen.io/chriddyp/pen/bWLwgP.css"]
_g_app = Dash(__name__, external_stylesheets=_g_external_stylesheets)
_g_country_indicators_df = pd.read_csv(
    "https://plotly.github.io/datasets/country_indicators.csv"
)

_g_common_config = {
    "graphOrientation": "vertical",
    "graphType": "Scatter2D",
    "lineBy": "Series",
    "smpLabelRotate": 90,
    "yAxisTitle": " ",
    "xAxisTitle": "Year",
    "xAxis2Show": False,
    "showLegend": False,
    "theme": "CanvasXpress",
    "titleAlign": "left",
    "titleScaleFontFactor": 0.65,
    "subtitleAlign": "left",
    "subtitleScaleFontFactor": 0.5,
}

_g_common_events = [
    CXEvent(
        id="mousemove",
        script="t.showInfoSpan(e, '<b>' + o.y.data[0][0] + '</b>: ' + o.y.data[0][1]);",
    ),
    CXEvent(id="mouseout", script=""),
]

# Establish the essential X axis chart configuration.  Only data and data specific configurations will be
# adjusted during the app's use.
_g_x_time_series_chart = CanvasXpress(
    render_to=None,
    config=_g_common_config,
    events=_g_common_events,
    height=225,
)

# Establish the essential Y axis chart configuration.  Only data and data specific configurations will be
# adjusted during the app's use.
_g_y_time_series_chart = CanvasXpress(
    render_to=None,
    config=_g_common_config,
    events=_g_common_events,
    height=225,
)

_g_app.layout = html.Div(
    [
        html.H1(
            "Mixed CanvasXpress and Plotly Graph Example",
            style={"width": "100%", "text-align": "center"},
        ),
        html.Hr(),
        html.Div(
            """
            This Dash application demonstrates a mixture of CanvasXpress and Plotly elements coordinating data
            exchange and UI refresh.  The CanvasXpress class is used for preparing CanvasXpress charts, and the 
            CXDashElementFactory is used to provide properly configured Dash elements for display.
            """
        ),
        html.Hr(),
        html.Div(
            [
                html.Div(style={"display": "inline-block", "width": "5%"}),
                html.Div(
                    [
                        html.Div(
                            [
                                dcc.Dropdown(
                                    _g_country_indicators_df["Indicator Name"].unique(),
                                    "Life expectancy at birth, total (years)",
                                    id="crossfilter-yaxis-column",
                                    style={"display": "inline-block", "width": "85%"},
                                ),
                                dcc.RadioItems(
                                    ["Linear", "Log"],
                                    "Linear",
                                    id="crossfilter-yaxis-type",
                                    style={"display": "inline-block", "width": "15%"},
                                ),
                            ],
                            style={"display": "inline-block", "width": "100%"},
                        ),
                    ],
                    style={"display": "inline-block", "width": "40%"},
                ),
                html.Div(style={"display": "inline-block", "width": "10%"}),
                html.Div(
                    [
                        html.Div(
                            [
                                dcc.Dropdown(
                                    _g_country_indicators_df["Indicator Name"].unique(),
                                    "Fertility rate, total (births per woman)",
                                    id="crossfilter-xaxis-column",
                                    style={"display": "inline-block", "width": "85%"},
                                ),
                                dcc.RadioItems(
                                    ["Linear", "Log"],
                                    "Linear",
                                    id="crossfilter-xaxis-type",
                                    style={"display": "inline-block", "width": "15%"},
                                ),
                            ],
                            style={"display": "inline-block", "width": "100%"},
                        ),
                    ],
                    style={"display": "inline-block", "width": "40%"},
                ),
                html.Div(style={"display": "inline-block", "width": "5%"}),
            ],
        ),
        html.Div(
            [
                html.Div(style={"display": "inline-block", "width": "2%"}),
                html.Div(
                    dcc.Slider(
                        _g_country_indicators_df["Year"].min(),
                        _g_country_indicators_df["Year"].max(),
                        step=None,
                        id="crossfilter-year--slider",
                        value=_g_country_indicators_df["Year"].max(),
                        marks={
                            str(year): str(year)
                            for year in _g_country_indicators_df["Year"].unique()
                        },
                    ),
                    style={"display": "inline-block", "width": "93%"},
                ),
                html.Div(style={"display": "inline-block", "width": "2%"}),
            ],
        ),
        html.Hr(),
        html.Div(
            [
                html.Div(
                    [
                        dcc.Graph(
                            id="crossfilter-indicator-scatter",
                            hoverData={"points": [{"customdata": "Japan"}]},
                        ),
                    ],
                    style={"display": "inline-block", "width": "50%"},
                ),
                html.Div(
                    [
                        html.Div(
                            [
                                html.Div(
                                    id="y-time-series",
                                    style={"display": "inline-block", "width": "100%"},
                                ),
                            ],
                        ),
                        html.Div(
                            [
                                html.Div(
                                    id="x-time-series",
                                    style={"display": "inline-block", "width": "100%"},
                                ),
                            ],
                        ),
                    ],
                    style={"display": "inline-block", "width": "50%"},
                ),
            ],
        ),
        html.Hr(),
    ]
)


@_g_app.callback(
    Output("crossfilter-indicator-scatter", "figure"),
    Input("crossfilter-xaxis-column", "value"),
    Input("crossfilter-yaxis-column", "value"),
    Input("crossfilter-xaxis-type", "value"),
    Input("crossfilter-yaxis-type", "value"),
    Input("crossfilter-year--slider", "value"),
)
def update_graph(
    xaxis_column_name, yaxis_column_name, xaxis_type, yaxis_type, year_value
):
    dff = _g_country_indicators_df[_g_country_indicators_df["Year"] == year_value]

    fig = px.scatter(
        x=dff[dff["Indicator Name"] == xaxis_column_name]["Value"],
        y=dff[dff["Indicator Name"] == yaxis_column_name]["Value"],
        hover_name=dff[dff["Indicator Name"] == yaxis_column_name]["Country Name"],
    )

    fig.update_traces(
        customdata=dff[dff["Indicator Name"] == yaxis_column_name]["Country Name"]
    )

    fig.update_xaxes(
        title=xaxis_column_name, type="linear" if xaxis_type == "Linear" else "log"
    )

    fig.update_yaxes(
        title=yaxis_column_name, type="linear" if yaxis_type == "Linear" else "log"
    )

    fig.update_layout(margin={"l": 40, "b": 40, "t": 10, "r": 0}, hovermode="closest")

    return fig


@_g_app.callback(
    Output("x-time-series", "children"),
    Input("crossfilter-indicator-scatter", "hoverData"),
    Input("crossfilter-xaxis-column", "value"),
    Input("crossfilter-xaxis-type", "value"),
)
def update_y_timeseries(hoverData, xaxis_column_name, axis_type):
    global _g_x_time_series_chart

    country_name = hoverData["points"][0]["customdata"]
    dff = _g_country_indicators_df[
        _g_country_indicators_df["Country Name"] == country_name
    ]
    dff = dff[dff["Indicator Name"] == xaxis_column_name]

    dff.drop(columns=["Country Name", "Indicator Name"], inplace=True)

    dff.Value.apply(lambda x: None if isnan(x) else x)
    dff.Value.apply(lambda x: x if axis_type == "Linear" else log(x))

    data = CXDataframeData(dff)
    data.profile.smps = ["Year", "Value"]
    data.profile.z = {
        "z": {"Series": ["Value" for i in range(dff["Value"].count())]},
    }

    _g_x_time_series_chart.data = data
    _g_x_time_series_chart.config.set_param("title", xaxis_column_name)
    _g_x_time_series_chart.config.set_param("subtitle", country_name)

    return CXElementFactory.render(_g_x_time_series_chart)


@_g_app.callback(
    Output("y-time-series", "children"),
    Input("crossfilter-indicator-scatter", "hoverData"),
    Input("crossfilter-yaxis-column", "value"),
    Input("crossfilter-yaxis-type", "value"),
)
def update_x_timeseries(hoverData, yaxis_column_name, axis_type):
    global _g_y_time_series_chart

    country_name = hoverData["points"][0]["customdata"]
    dff = _g_country_indicators_df[
        _g_country_indicators_df["Country Name"] == country_name
    ]
    dff = dff[dff["Indicator Name"] == yaxis_column_name]

    dff.drop(columns=["Country Name", "Indicator Name"], inplace=True)

    dff.Value.apply(lambda x: None if isnan(x) else x)
    dff.Value.apply(lambda x: x if axis_type == "Linear" else log(x))

    data = CXDataframeData(dff)
    data.profile.z = {
        "z": {"Series": ["Series" for i in range(dff["Value"].count())]},
    }

    _g_y_time_series_chart.data = data
    _g_y_time_series_chart.config.set_param("title", yaxis_column_name)
    _g_y_time_series_chart.config.set_param("subtitle", country_name)

    return CXElementFactory.render(_g_y_time_series_chart)


if __name__ == "__main__":
    _g_app.run_server(debug=False)



"""
This file can be executed to render the reproducible JSON files located at
`[project]/tutorials/reproducible_json/*.json` into tutorials for general use.
"""

import json
import os
from pathlib import Path
from typing import List

from canvasxpress.util.example.generator import (
    generate_canvasxpress_code_from_json_file,
)

JSON_DIR_PATH = f"{os.getcwd()}/../../../tutorials/reproducible_json/"
JUPYTER_TEMPLATE_PATH = (
    f"{os.getcwd()}/../../../canvasxpress/util/" f"example/template_tutorials.ipynb"
)
JUPYTER_EXAMPLES_DIR_PATH = (
    f"{os.getcwd()}/../../../tutorials/notebook/" f"cx_site_chart_examples/"
)


def get_json_file_paths() -> List[str]:
    """
    Returns a list of all reproducible JSON files tracked for tutorials.
    :returns: `list[str]`
        The file paths as a list of strings.
    """
    json_files = list()
    for file in os.listdir(JSON_DIR_PATH):
        if file.endswith(".json"):
            json_files.append(os.path.join(JSON_DIR_PATH, file))
    return sorted(json_files)


def get_type_from_filename(file_name: str) -> str:
    """
    Returns the type of chart from a reproducible JSON filename.
    :param file_name: `str`
        The name of the file without parent path.
    :returns: `str`
        The name of the chart (e.g., bar) or an empty string.
    """
    assembled_type = ""
    started_type = False
    for name_char in file_name.replace(".json", "")[::-1]:
        if not started_type and name_char.isnumeric():
            continue

        else:
            started_type = True
            assembled_type += name_char

    return assembled_type[::-1]


def get_index_from_filename(file_name: str) -> str:
    """
    Returns the index of chart from a reproducible JSON filename.
    :param file_name: `str`
        The name of the file without parent path.
    :returns: `str`
        The index of the chart (e.g., 1) or an empty string.
    """
    assembled_index = ""
    for name_char in file_name.replace(".json", "")[::-1]:
        if name_char.isnumeric():
            assembled_index += name_char

        else:
            break

    return assembled_index[::-1]


def create_jupyer_template_text(
    chart_type: str, chart_index: str, chart_code: str
) -> str:
    """
    Generates the text for a Jupyter Notebook example given a chart's type,
    index, and code.
    :param: chart_type: `str`
        The type text (e.g., bar) for the chart.
    :param chart_index: `str`
        The index text (e.g., 1) for the chart.
    :param chart_code: `str`
        The chart source code.
    :returns: `str`
        The text for the full example and instruction.
    """
    with open(JUPYTER_TEMPLATE_PATH, "r") as template_file:
        example_text = template_file.read()
        example_text = example_text.replace("@type@", chart_type)
        example_text = example_text.replace("@index@", chart_index)

        ipython_json = json.loads(example_text)
        for line in chart_code.splitlines():
            candidate = line

            # Convert render statement to explicit output
            if "display.render()" in candidate:
                candidate = candidate.replace(
                    "display.render()",
                    f'display.render(output_file="{chart_type}_{chart_index}.html")',
                )

            # Add the source line to the document
            ipython_json["cells"][1]["source"].append(candidate + "\n")

        ipython_text = json.dumps(ipython_json)
        return ipython_text


if __name__ == "__main__":
    json_paths = get_json_file_paths()
    for json_path in json_paths:
        try:
            json_name = Path(json_path).name

            chart_type = get_type_from_filename(json_name)
            chart_index = get_index_from_filename(json_name)

            jupyter_notebook_content = create_jupyer_template_text(
                chart_type,
                chart_index,
                generate_canvasxpress_code_from_json_file(
                    json_path, document_jupyter_render=True
                ),
            )

            example_file_name = f"{chart_type}_{chart_index}.ipynb"
            example_file_path = str(
                Path(JUPYTER_EXAMPLES_DIR_PATH).joinpath(example_file_name)
            )
            with open(example_file_path, "w") as example_file:
                example_file.write(jupyter_notebook_content)

        except Exception as e:
            print(f"Cannot process file: {json_path}")
            print(f"Exception: {e}")



import json
from copy import copy, deepcopy
from functools import total_ordering
from typing import List

from canvasxpress.data.convert import CXJavascriptConvertable
from canvasxpress.js.function import CXEvent


@total_ordering
class CXEvents(CXJavascriptConvertable):
    """
    CXEvents represents a Javascript script that can be associated with a
    CanvasXpress object.

    For example, when defining a CanvasXpress object Javascript could be
    added for reactive feedback via:

    ```python
    chart_events = CXEvents(
        CXEvent(
            "click",
            '''
            var s = 'click on var ' + o.y.vars[0] + ' and smp ' + o.y.smps[0];
            t.showInfoSpan(e, s);
            '''
        )
    )
    chart = CanvasXpress(
        events=chart_events
    )
    ```

    Also see the [CanvasXpress documentation](https://www.canvasxpress.org/docs.html#events)
    or `CXEvent` for additional information.
    """

    __events: List[CXEvent] = list()
    """
    The `CXEvent` objects tracked by this instance.
    """

    @property
    def events(self) -> List[CXEvent]:
        """
        Provides a non-associated list of the associated CXEvents.
        :returns: `List[CXEvent]` A list of zero or more CXEvent objects.
        """
        return copy(self.__events)

    def has(self, event: CXEvent) -> bool:
        """
        Indicates if the `CXEvent` is a member.
        :param event: The `CXEvent` to consider.
        :returns: `bool` True if `event` is a member.
        """
        return event in self.events

    def add(self, event: CXEvent, unique: bool = True) -> None:
        """
        Adds the specified CXEvent.  If the CXEvent must be unique then an Error
        is raised if an react is already presenbt with the same ID.
        :param event:
            `CXEvent` The event to add to the collection.  Cannot be `None`.
        :param unique:
            `bool` True if `event` must not already be a part of the collection.
        """
        if not event:
            raise TypeError("event cannot be None.")

        if not isinstance(event, CXEvent):
            raise TypeError("react must be of type CXEvent.")

        if unique and self.has(event):
            raise ValueError(
                f"A CXEvent of ID '{event.id}' is already part "
                f"of this CXEvent object."
            )

        self.__events.append(event)

    def remove(self, event: CXEvent) -> bool:
        """
        Removes the specified object from the list.
        :param event: The CXEvent object to remove from the list if it is
            already included.
        :returns: True if the CXEvent was removed.  False indicates that the
            object was not a member.
        """
        if self.has(event):
            for associated_event in self.__events:
                if associated_event == event:
                    self.__events.remove(event)
            return True

        else:
            return False

    def render_to_dict(self) -> dict:
        """
        Provides a dict with each js properly formatted as JS within.
        :returns: `dict`
        Given:
        ```python
        event1 = CXEvent("f1", "x = 0")
        event2 = CXEvent("f2", "x = 1")
        events = CXEvents(event1, event2)
        functions = events.render_to_dict()
        ```
        Then the value of `functions` would be:
        ```python
        {
            "f1": function(o, e, t){x = 0},
            "f2": function(o, e, t){x = 1}
        }
        ```
        """

        events = dict()
        for event in self.events:
            events[event.id] = event.render_to_js()

        return events

    def render_to_js(self) -> str:
        """
        Converts the object into HTML5 complant script.
        :returns: 'str'
        Given:
        ```python
        event1 = CXEvent("f1", "x = 0")
        event2 = CXEvent("f2", "x = 1")
        events = CXEvents(event1, event2)
        functions = events.render_to_js()
        ```
        Then the value of `functions` would be:
        ```text
        {
            'f1': 'function(o, e, t){x = 0}',
            'f2': 'function(o, e, t){x = 1}',
        }
        ```
        """
        events = dict()
        for event in self.events:
            events[event.id] = f"js_{event.id}"

        html = json.dumps(events).replace('"', "'")
        for event in self.events:
            html = html.replace(f"'js_{event.id}'", event.render_to_js())

        return html

    def __init__(self, *events):
        """
        Initializes a new CXEvents object.
        :param events:
            A multiple value parameter by which zero or more `CXEvent` objects
            may be provided.  Also see `add()` for how individual objects are
            processed.

            For example:
        ```python
        event1 = CXEvent("f1", "x = 0")
        event2 = CXEvent("f2", "x = 1")
        events = CXEvents(event1, event2)
        ```
        """
        super().__init__()

        self.__events = list()
        if events:
            for event in events:
                self.add(event)

    def __copy__(self):
        """
        *copy* constructor.  Returns the `CXEvent` objects within a new `CXEvents`
        object.
        """
        return CXEvents(*self.events)

    def __deepcopy__(self, memo):
        """
        *deepcopy* constructor.  Returns a deep copy of `CXEvent` objects within
        a new `CXEvents` object.
        """
        return CXEvents(*([deepcopy(event) for event in self.events]))

    def __lt__(self, other: "CXEvents"):
        """
        *less than* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXEvent` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXEvents` object then False
            <li> If `other` is a `CXEvents` object then True of all `CXEvent`
                objects are also less than the events tracked by `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXEvents):
            return False

        else:
            if (len(self.events) + len(other.events)) == 0:
                return False

            if len(self.events) == len(other.events):
                for event in self.events:
                    for oevent in other.events:
                        if not event < oevent:
                            return False
                return True

            else:
                return len(self.events) < len(other.events)

    def __eq__(self, other: "CXEvents"):
        """
        *equals* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXEvent` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXEvents` object then False
            <li> If `other` is a `CXEvents` object then True of all `CXEvent`
                objects are also equal to the events tracked by `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXEvents):
            return False

        else:
            if len(self.events) == len(other.events):
                for event in self.events:
                    for oevent in other.events:
                        if not event == oevent:
                            return False
                return True

            else:
                return len(self.events) == len(other.events)

    def __str__(self) -> str:
        """
        *str* function.  Converts the CXEvents object into a JSON list of
        `CXEvent` objects also converted into JSON representations.
        :returns" `str` JSON form of the collection.
        """
        return json.dumps(self.render_to_dict())

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXEvents object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        event_rep_list = ", ".join([repr(event) for event in self.events])
        rep_candidate = f"CXEvents(" f"{event_rep_list}" f")"
        return rep_candidate



import uuid
from pathlib import Path
from typing import Any, Union, List

import htmlmin
from IPython.display import display, HTML, Code
from bs4 import BeautifulSoup

from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.base import CXRenderable

_cx_iframe_padding = 50

_cx_intermixed_header = """
<html>
    <head>
        <meta charset="UTF-8">
        <link 
                href='@css_url@' 
                rel='stylesheet' 
                type='text/css'
        />
        <script 
                src='@js_url@' 
                type='text/javascript'>
        </script>
    </head>
    <body></body>
</html>
"""

_cx_js_intermixed_template = """
<script type="text/javascript">
    @code@
</script>
"""

_cx_html_intermixed_template = """
<html>
    <head>
        <meta charset="UTF-8">
    </head>
    <body>
        @canvasxpress_license@
        @canvases@
        @js_functions@
    </body>
</html>
"""

_nb_iframe_template = "data:text/html,@html@"


class CXNoteBook(CXRenderable):
    """
    CXNoteBook is a `CXRenderable` that renders `CanvasXpress` objects into
    `IPython` containers (Jupyter Notebooks).
    """

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXNoteBook` object.
        :praram cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        super().__init__(*cx)

    @classmethod
    def display_canvasxpress_header(cls):
        css_url = CanvasXpress.css_library_url()
        js_url = CanvasXpress.js_library_url()

        header_html_text = _cx_intermixed_header.replace("@css_url@", css_url).replace(
            "@js_url@", js_url
        )

        display(
            HTML(
                data=header_html_text,
            ),
        )

    def display_debug_code(self, code: str):
        minified_code = htmlmin.Minifier().minify(code)
        pretty_code = BeautifulSoup(minified_code, "html.parser").prettify()
        display(
            Code(
                data=pretty_code,
                language="html",
            ),
        )

    def get_chart_display_code(self, columns: int) -> str:
        render_targets = list()

        if self.canvas is None:
            pass

        elif isinstance(self.canvas, CanvasXpress):
            render_targets.append(self.canvas)

        else:
            render_targets.extend(self.canvas)

        used_render_targets = list()
        for target in render_targets:
            original_render_target = target.render_to
            if original_render_target in used_render_targets:
                target.render_to = (
                    original_render_target + "_" + str(uuid.uuid4()).replace("-", "_")
                )

            used_render_targets.append(target.render_to)

        render_targets.reverse()

        html_parts = [target.render_to_html_parts() for target in render_targets]

        canvases = [part["cx_canvas"] for part in html_parts]
        if len(canvases) < columns:
            columns = len(canvases)

        functions = [part["cx_js"] for part in html_parts]

        cx_license = ""
        for part in html_parts:
            if part.get("cx_license"):
                cx_license = part["cx_license"]
                break

        iframe_width = 0
        iframe_height = 0
        chart_count = len(canvases)

        canvas_table = '<div class="d-flex flex-column">'

        while chart_count > 0:
            candidate_width = 0
            candidate_height = 0

            canvas_table += '<div class="d-flex flex-row">'
            for c in range(columns):
                canvas_table += '<div class="p-2">'
                if chart_count > 0:
                    canvas_table += canvases[chart_count - 1]

                    candidate_width += render_targets[chart_count - 1].width
                    if render_targets[chart_count - 1].height > candidate_height:
                        candidate_height = render_targets[chart_count - 1].height

                canvas_table += "</div>"
                chart_count = chart_count - 1

            canvas_table += "</div>"

            if candidate_width > iframe_width:
                iframe_width = candidate_width
            iframe_height += candidate_height

        canvas_table += "</div>"

        iframe_width += _cx_iframe_padding
        iframe_height += _cx_iframe_padding

        js_functions = "\n".join(
            [_cx_js_intermixed_template.replace("@code@", fx) for fx in functions]
        )
        html_text = (
            _cx_html_intermixed_template.replace("@canvases@", canvas_table)
            .replace("@canvasxpress_license@", cx_license)
            .replace("@js_functions@", js_functions)
        )

        return html_text

    def display_charts(self, code: str, output_file: str):
        try:
            if output_file is not None:
                file_path_candidate = str(output_file)
                file_path = Path(file_path_candidate)
                if file_path.is_dir():
                    file_path = file_path.joinpath(f"cx_{str(uuid.uuid4())}.html")

                with open(str(file_path), "w") as render_file:
                    render_file.write(code)

        except Exception as e:
            raise RuntimeError(f"Cannot create output file: {e}")

        display(
            HTML(
                data=code,
            ),
        )

    def render(self, **kwargs: Any):
        """
        Renders the associated CanvasXpress object appropriate for display in
        an IPython (e.g., Jupyter NoteBook/Lab) environment.  Charts cannot
        have the same name, so render_to will be updated with a uuid for each
        conflicting chart.
        :param kwargs: `Any`
            * Supports `columns` for any positive `int` of `1` or greater, with a
              default value of `1`.  Values less that `1` are ignored.  `columns`
              indicates how many charts should be rendered horizontally in the
              Jupyter Notebook if more than one chart is being tracked.
            * Supports `output_file` as a string for a path at which the output
              should be saved.  If a file exists at the specified path then
              it will be overwritten.  This permits Jupyter sessions to render
              output that is saved and accessible in later sessions.
            * Supports `debug` for displaying the output source.  True indicates
              that the HTML code shall be displayed prior to the parsed output.
              Default is False.
        """
        debug_output_arg = kwargs.get("debug")
        debug_output = bool(debug_output_arg) if debug_output_arg is not None else False

        columns_arg = int(kwargs.get("columns", 1))
        columns = columns_arg if columns_arg > 0 else 1

        output_file_arg = kwargs.get("output_file")
        output_file = (
            output_file_arg
            if output_file_arg is not None and isinstance(output_file_arg, str)
            else None
        )

        code = self.get_chart_display_code(columns)

        try:
            # self.display_canvasxpress_header()
            self.display_charts(code, output_file)
            if debug_output:
                self.display_debug_code(code)

        except Exception as e:
            raise RuntimeError(f"Cannot create output cell: {e}")



"""
The CanvasXpress package provides Python friendly management of the Javascript-
based [CanvasXpress](https://www.canvasxpress.org) library.  For an overview
and detailed instructions about CanvasXpress specifically please visit the site.
"""



import os
import subprocess
import json
from typing import Any, Union, List
from tempfile import gettempdir
from pathlib import Path

from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.base import CXRenderable
from canvasxpress.render.json import CXJSON

PNG_IMAGE: str = "png"
SVG_IMAGE: str = "svg"

MAX_NODE_WAIT_SECONDS: int = 60


def install_cx_in_nodejs() -> None:
    """
    Installs the canvasxpress-cli NodeJS package if it is not available.
    """
    try:
        availability_status = subprocess.run(
            ["npm", "ls", "canvasxpress-cli"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            timeout=MAX_NODE_WAIT_SECONDS,
        )
        if availability_status.returncode != 0:
            installation_status = subprocess.run(
                ["npm", "install", "canvasxpress-cli", "--save"],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                timeout=MAX_NODE_WAIT_SECONDS,
            )
            if installation_status.returncode != 0:
                raise RuntimeError(
                    f"Image rendering is unavailable. "
                    f"Cannot install canvasxpress NodeJS package: {installation_status.stderr.decode('utf-8')}"
                )

    except subprocess.TimeoutExpired as te:
        raise RuntimeError(
            f"Image rendering is unavailable. " f"Cannot interact with Node: {str(te)}"
        )


def nodejs_modules_path() -> Path:
    """
    Returns the path to the NodeJS' modules installation.
    :returns: `str`
        The path for the NodeJS' modules installation.
    """
    install_cx_in_nodejs()
    search_path = Path(os.getcwd())
    while search_path != "/":
        test_path = Path(search_path, "node_modules")
        if test_path.exists():
            return test_path
        else:
            if search_path == search_path.root:
                raise RuntimeError("node_modules is not available.")
            else:
                search_path = search_path.parent


CX_NODEJS_PATH: Path = None


def get_nodejs_path() -> str:
    global CX_NODEJS_PATH
    if CX_NODEJS_PATH is None:
        CX_NODEJS_PATH = nodejs_modules_path() / "canvasxpress-cli/bin/canvasxpress"

    return CX_NODEJS_PATH


def render_html_as_image(
    url: str,
    format: Union[str, list] = PNG_IMAGE,
    width: Union[int, None] = None,
    height: Union[int, None] = None,
) -> list:
    """
    Renders a Web page with CanvasXpress chart declarations into one image per chart.
    :param url: `str`
        The URL to the HTML.  Can be file:// or http[s]://.
    :param format: `str`
        A `str` or list indicating whether PNG or SVG images should be produced.
    :returns: `list`
        A `list[dict]` of image data, if any.
    """
    formats = format if isinstance(format, list) else [format]

    if not (isinstance(width, int) or width is None):
        raise ValueError("width must be an int or None.")
    elif width is None:
        width_text = ""
    else:
        width_text = f" -x {width}"

    if not (isinstance(height, int) or height is None):
        raise ValueError("height must be an int or None.")
    elif height is None:
        height_text = ""
    else:
        height_text = f" -y {height}"

    rendered_images: list = []
    for image_format in formats:
        if image_format not in [PNG_IMAGE, SVG_IMAGE]:
            raise ValueError(
                "format must be one of PNG_IMAGE or SVG_IMAGE, or a list of each."
            )

        work_dir = Path(gettempdir()) / "canvasxpress-python"
        work_dir.mkdir(exist_ok=True)
        work_image_path = work_dir

        try:
            result = subprocess.run(
                [
                    f"{get_nodejs_path()} {image_format}{width_text}{height_text} -i {url} -o {work_image_path}"
                ],
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                timeout=MAX_NODE_WAIT_SECONDS,
            )
            if result.returncode == 0:

                image_file_options = list(work_image_path.glob(f"**/*.{image_format}"))
                for image_file_path in image_file_options:
                    with open(Path(image_file_path), "rb") as image_file:
                        image: bytes = image_file.read()
                    image_file_path.unlink()
                    rendered_images.append(
                        {
                            "id": os.path.basename(
                                os.path.splitext(image_file_path)[0]
                            ),
                            "image": {
                                "binary": image,
                                "format": image_format,
                            },
                        }
                    )

        except subprocess.TimeoutExpired as te:
            pass

    return rendered_images


class CXImage(CXRenderable):
    """
    CXPng is a `CXRenderable` that renders `CanvasXpress` objects into PNG images without the need for a Web session
    at the client code level.  NodeJS must be installed in the system, and the Python process must have permission
    to interact with NodeJS.
    """

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXNoteBook` object.
        :param cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        super().__init__(*cx)

    def render(self, **kwargs: Any) -> list:
        """
        Renders the associated CanvasXpress object appropriate to create PNGs.
        :param kwargs: `Any`
            * `format`: Accepts a `str` or `list[str]` with the values `png`, or `svg`.  Each tracked CanvasXpress
              object will be rendered into the specified image formats.  If not provided, then `png` is assumed.
        :returns: `list{dict}`
            A list `dict`, each containing a PNGs or SVG and the `render_to` ID of the corresponding CanvasXpress.
        """
        format_arg = kwargs.get("format", [PNG_IMAGE])
        formats = format_arg if isinstance(format_arg, list) else [format_arg]

        json_renderer = CXJSON(self.canvas)
        reproducible_jsons = json_renderer.render()

        rendered_images: list = []

        for image_format in formats:
            if image_format not in [PNG_IMAGE, SVG_IMAGE]:
                raise ValueError(
                    "format must be one of PNG_IMAGE or SVG_IMAGE, or a list of each."
                )

            for reproducible_json in reproducible_jsons:
                work_dir = Path(gettempdir()) / "canvasxpress-python"
                work_dir.mkdir(exist_ok=True)
                work_json_path = Path(work_dir, "cx_data.json")
                work_image_path = work_dir

                with open(work_json_path, "w") as json_temp_file:
                    json_temp_file.write(reproducible_json)

                result = subprocess.run(
                    [
                        f"{get_nodejs_path()} {image_format} -i {work_json_path} -o {work_image_path}",
                    ],
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    timeout=MAX_NODE_WAIT_SECONDS,
                )
                work_json_path.unlink()
                if result.returncode == 0:

                    image_file_options = list(
                        work_image_path.glob(f"**/*.{image_format}")
                    )
                    for image_file_path in image_file_options:
                        with open(Path(image_file_path), "rb") as image_file:
                            image: bytes = image_file.read()
                        image_file_path.unlink()
                        rendered_images.append(
                            {
                                "id": json.loads(reproducible_json).get(
                                    "renderTo", "anonymous"
                                ),
                                "image": {
                                    "binary": image,
                                    "format": image_format,
                                },
                            }
                        )

        return rendered_images



# AUTO GENERATED FILE - DO NOT EDIT

import typing  # noqa: F401
from typing_extensions import TypedDict, NotRequired, Literal # noqa: F401
from dash.development.base_component import Component, _explicitize_args

ComponentSingleType = typing.Union[str, int, float, Component, None]
ComponentType = typing.Union[
    ComponentSingleType,
    typing.Sequence[ComponentSingleType],
]

NumberType = typing.Union[
    typing.SupportsFloat, typing.SupportsInt, typing.SupportsComplex
]


class CXDashElement(Component):
    """A CXDashElement component.
CXDashElement implements a Plotly Dash integration of CanvasXpress React.
Properties are defined for use by the CanvasXpress class to update
CanvasXpress aspects such as data, config, and sizing.

Keyword arguments:

- id (string; required):
    The ID of the element for use in function calls and element
    identification.

- after_render (string; optional):
    The events functions for increased reactivity.

- config (string; optional):
    The configuration JSON dictating formatting and content
    management.

- css_url (string; optional):
    The CSS CDN URL that should be used for CanvasXpress
    functionality.

- data (string; optional):
    The data JSON, generally in the XYZ format.

- events (string; optional):
    The events functions for increased reactivity.

- height (string; optional):
    The element height.

- js_url (string; optional):
    The Javascript URL that should be used for CanvasXpress
    functionality.

- width (string; optional):
    The element width."""
    _children_props: typing.List[str] = []
    _base_nodes = ['children']
    _namespace = 'cxdash'
    _type = 'CXDashElement'


    def __init__(
        self,
        id: typing.Optional[typing.Union[str, dict]] = None,
        data: typing.Optional[str] = None,
        config: typing.Optional[str] = None,
        events: typing.Optional[str] = None,
        after_render: typing.Optional[str] = None,
        js_url: typing.Optional[str] = None,
        css_url: typing.Optional[str] = None,
        width: typing.Optional[str] = None,
        height: typing.Optional[str] = None,
        **kwargs
    ):
        self._prop_names = ['id', 'after_render', 'config', 'css_url', 'data', 'events', 'height', 'js_url', 'width']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'after_render', 'config', 'css_url', 'data', 'events', 'height', 'js_url', 'width']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        for k in ['id']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')

        super(CXDashElement, self).__init__(**args)

setattr(CXDashElement, "__init__", _explicitize_args(CXDashElement.__init__))



"""
The data package provides functionality for integrating different kinds of data
structures and sources into a CanvasXpress object.  A balance is provided
between convenience functionality and reasonable external preparation, such as
what can be performed via `pandas`.
"""



import json
from copy import deepcopy
from functools import total_ordering
from typing import Union

import requests
from deepdiff import DeepDiff

from canvasxpress.data.base import CXKeyPairData


@total_ordering
class CXDictData(CXKeyPairData):
    """
    A CXData class dedicated to processing Python dict-structured data.
    """

    __data: dict = dict()
    """
    The data managed by an object of this class.
    """

    @property
    def data(self) -> dict:
        """
        Provides a reference to the dict tracked by the object.
        :returns: `dict`
            The associated dictionary, with zero or more keys as appropriate.
        """
        return self.__data

    @data.setter
    def data(self, value: dict) -> None:
        """
        Sets the data associated with the object.
        :param value: `dict`
            The dictionary to be tracked by the object.  `None` will result in
            an empty dict.  A deep copy will be made of a valid `CXDict` or
            `dict` provided.
        """
        if value == None:
            self.__data = dict()

        elif not type(value) in [dict, CXDictData]:
            raise TypeError("value must be type dict or compatible.")

        elif isinstance(value, CXDictData):
            self.__data = deepcopy(value.data)

        else:
            self.__data = deepcopy(value)

    def get_raw_dict_form(self) -> dict:
        """
        Provides a simple dict perspective of the data with no metadata or other
        contextual transformations performed.  For example, if the data is
        natively in `dict` form then it would be passed-through with no
        modification or enhancement.

        This implementation provides matrix data formatted in a `dict` object
        with `DataFrame.to_dict('split')` behaviour.

        :returns: `dict`
            The `dict` perspective of the data with as little modification or
            interpretation as is reasonable.
        """
        return deepcopy(self.data)

    def render_to_dict(self, **kwargs) -> dict:
        """
        Provides a dict representation of the data.
        :returns: `dict`
            The data in `dict` form.
        """
        return self.get_raw_dict_form()

    def __init__(self, data: Union[dict, None] = None) -> None:
        """
        Initializes the CXData object with data.  Only dict or compatible data
        types are accepted.
        :param data: `Union[dict, None]`
            `None` to initialize with an empty dictionary, or a `dict`-like
            object to assign mapped data.
        """
        super().__init__(data)
        self.data = data

    def __copy__(self) -> "CXDictData":
        """
        *copy constructor* that returns a copy of the CXDictData object.
        :returns: `CXDictData` A copy of the wrapping object.
        """
        return self.__class__(self.data)

    def __deepcopy__(self, memo) -> "CXDictData":
        """
        *deepcopy constructor* that returns a copy of the CXDictData object.
        :returns: `CXDictData` A copy of the wrapping object and deepcopy of
            the tracked data.
        """
        return self.__class__(deepcopy(self.data))

    def __lt__(self, other: "CXDictData") -> bool:
        """
        *less than* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXDictData` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXDictData` object then False
            <li> If `other` is a `CXDictData` object then True of all
                `CXDictData` objects are also less than the data tracked by
                `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXDictData):
            return False

        else:
            delta: dict = DeepDiff(self.data, other.data, ignore_order=True)
            other_added: int = len(delta.get("dictionary_item_added", []))
            other_removed: int = len(delta.get("dictionary_item_removed", []))

            if (other_added - other_removed) == 0:
                for skey in self.data.keys():
                    if not skey in other.data.keys():
                        for okey in other.data.keys():
                            if skey < okey:
                                return True

                    elif self.data[skey] < other.data[skey]:
                        return True

                return False

            else:
                return (other_added - other_removed) > 0

    def __eq__(self, other: "CXDictData") -> bool:
        """
        *equals* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXDictData` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXDictData` object then False
            <li> If `other` is a `CXDictData` object then True of all
                `CXDictData` objects are also equal to the data tracked by
                `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXDictData):
            return False

        else:
            delta: dict = DeepDiff(self.data, other.data, ignore_order=True)
            other_added: int = len(delta.get("dictionary_item_added", []))
            other_removed: int = len(delta.get("dictionary_item_removed", []))

            if (other_added - other_removed) == 0:
                for skey in self.data.keys():
                    if not skey in other.data.keys():
                        return False

                    elif self.data[skey] != other.data[skey]:
                        return False

                return True

            else:
                return False

    def __str__(self) -> str:
        """
        *str* function.  Converts the CXDictData object into a JSON
        representation.
        :returns" `str` JSON form of the `CXDictData`.
        """
        return json.dumps(self.data)

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXDictData object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        return f"CXDictData(data={json.dumps(self.data)})"


class CXJSONData(CXDictData):
    """
    A CXData class dedicated to processing JSON data.
    """

    @property
    def json(self) -> str:
        """
        Provides a copy of the JSON tracked by the object.
        :returns: `str`
            The associated JSON, with zero or more keys as appropriate.
        """
        return str(self)

    @json.setter
    def json(self, value: Union[dict, str]) -> None:
        """
        Sets the data associated with the object.
        :param value: `str`
            The JSON to be tracked by the object.  `None` will result in
            an empty JSON.  If `value` is URL beginning with *http/s*
            then `json` will attempt to download the data.
        """
        if isinstance(value, str) and value.lower().startswith("http"):
            try:
                result = requests.get(value, allow_redirects=True)
                self.data = result.json()

            except Exception as e:
                raise ValueError(
                    "Detected URL but cannot access JSON data at endpoint."
                )

        else:
            self.data = value

    @CXDictData.data.setter
    def data(self, value: Union[dict, str]) -> None:
        if isinstance(value, CXJSONData):
            CXDictData.data.fset(self, value.data)

        elif isinstance(value, str):
            candidate = json.loads(value)
            CXDictData.data.fset(self, candidate)

        else:
            CXDictData.data.fset(self, value)

    def __init__(self, data: Union[dict, str, None] = None) -> None:
        """
        Initializes the CXData object with data.  Only dict or compatible data
        types are accepted.
        :param data: `Union[dict, str, None]`
            `None` to initialize with an empty JSON, or a JSON/`dict`-like
            object to assign mapped data.
        """
        super().__init__(data)

    def __copy__(self) -> "CXJSONData":
        """
        *copy constructor* that returns a copy of the CXDictData objct.
        :returns: `CXDictData` A copy of the wrapping object.
        """
        return CXJSONData(self.data)

    def __deepcopy__(self, memo) -> "CXJSONData":
        """
        *deepcopy constructor* that returns a copy of the CXJSONData object.
        :returns: `CXJSONData` A copy of the wrapping object and deepcopy of
            the tracked data.
        """
        return CXJSONData(self.data)

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXJSONData object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        return f"CXJSONData(data={str(self.data)})"



import json
from copy import deepcopy
from functools import total_ordering
from typing import List, Any, Union

from canvasxpress.config.type import (
    CXConfig,
    CXString,
    CXInt,
    CXFloat,
    CXBool,
    CXList,
    CXDict,
    CXNone,
    CXGraphWeight,
)
from canvasxpress.data.convert import CXDictConvertable, CXListConvertable


@total_ordering
class CXConfigs(CXDictConvertable, CXListConvertable):
    """
    CXConfigs provides support for addressing a collection of `CXConfig` values.
    """

    __configs: List[CXConfig] = None
    """
    The `CXConfig` objects associated with this collection.
    """

    def __init__(self, *configs: Union[CXConfig, tuple, dict, list]):
        """
        Initializes a new `CXConfigs` object with zero or more `CXConfig`
        objects.

        Example:
        ```python
        configs = CXConfigs(
            CXString("colorScheme", "ColorSpectrum"),
            ("lineType", "spline"),
            { "objectColorTransparency": 0.3 }
        )
        ```

        :param configs: `Union[CXConfig, tuple, dict], ...`
            A list of zero or more `CXConfig` objects to associate.
        """
        self.__configs: List[CXConfig] = list()
        for config in configs:
            self.add(config)

    def remove(self, label: str) -> Union[CXConfig, None]:
        """
        Removes the CXConfig if found, and if found the removed config is
        provided.
        :param label: 'str`
            The label of the CXConfig to remove.
        :returns: `Union[CXConfig, None]`
            If a CXConfig is removed then it is returned, otherwise None.
        """
        candidate = None
        for config in self.configs:
            if str(label) == config.label:
                candidate = config
                self.configs.remove(config)
                break

        return candidate

    def add(self, config: Union[CXConfig, tuple, dict, list]) -> "CXConfigs":
        """
        Adds the specified configuration to the collection.  This method
        supports chaining for efficient additions of `CXConfig` objects.

        Example:
        ```python
        configs = CXConfigs()
        configs \
            .add(CXString("colorScheme", "ColorSpectrum") \
            .add(("lineType", "spline")) \
            .add({ "objectColorTransparency": 0.3 })
        ```

        :param config: `Union[CXConfig, tuple, dict, list]`
            The `CXConfig` to associate.  Cannot be `None`.  `tuple` an d`list`
            config values are expected to be two elements in length, with the
            first representing the label and the second representing the value.
            The label portion will be converted to a string using `str`.
        """
        if config is None:
            raise ValueError("configs cannot be None.")

        elif isinstance(config, dict):
            for param in config.keys():
                self.set_param(str(param), config.get(param))

        elif isinstance(config, (tuple, list)):
            if len(config) > 0:
                if isinstance(config[0], (tuple, list)):
                    for item in config:
                        self.add(item)

                else:
                    if len(config) < 2:
                        raise ValueError(
                            "list and tuple config representations of "
                            "configuration parameters must have two elements: "
                            "the first is the label, and the second is the "
                            "value."
                        )
                    self.set_param(str(config[0]), config[1])

        elif isinstance(config, CXConfig):
            if config not in self.__configs:
                self.__configs.append(config)

        else:
            raise TypeError("configs must be a type of CXConfig.")

        return self

    def get_param(self, label: str) -> Union[CXConfig, None]:
        """
        Provides the CXConfig with the indicated label.
        :param label: `str`
            The name of the congig to find.
        :returns: `Union[CXConfig, None]`
            The CXConfig or None if such a labelled item is not associated.
        """
        candidate = None
        for config in self.configs:
            if config.label == label:
                candidate = config
                break

        return candidate

    def set_param(self, label: str, value: Any) -> "CXConfigs":
        """
        Adds a parameter to the configs.  Attempts to infer the kind of param to
        add, and if a type can be deduced then an appropriate CXConfig is used.
        If a type cannot be inferred the a text type is assumed. This method
        supports chaining for efficient additions of `CXConfig` objects.

        Example:
        ```python
        configs = CXConfigs()
        configs \
            .set_param("1", "rgb(3, 172, 198)") \
            .set_param("2", 2) \
            .set_param("3", True)
        ```

        :param value: `Any`
            The parameter to infer and associate.  Cannot be `None`.  Defaults
            to `str` if the type cannot otherwise be deduced.
        """
        if label is None:
            raise ValueError("Label cannot be None.")

        else:
            existing_config_used = False
            for config in self.configs:
                if config.label == label:
                    if not isinstance(config.value, type(value)):
                        raise ValueError(
                            f"CXConfig {repr(config)} is already a member and"
                            f" has a different type than what is provided."
                            f" Remove the existing CXConfig object first or use"
                            f" the same type."
                        )
                    else:
                        config.value = value
                        existing_config_used = True
                        break

            if not existing_config_used:
                value_type = type(value)
                if value is None:
                    candidate = CXNone(label, value)
                elif value_type is int:
                    candidate = CXInt(label, value)
                elif value_type is float:
                    candidate = CXFloat(label, value)
                elif value_type is bool:
                    candidate = CXBool(label, value)
                elif value_type is dict:
                    candidate = CXDict(label, value)
                elif value_type is list:
                    if CXGraphWeight.is_graph_weight_list(label, value):
                        candidate = CXGraphWeight(label, value)
                    else:
                        candidate = CXList(label, value)
                elif value_type is set:
                    set_persona: set = value
                    candidate = CXList(label, list(set_persona))
                else:
                    candidate = CXString(label, value)

                self.add(candidate)

            return self

    @property
    def configs(self) -> List[CXConfig]:
        """
        Provides access to the list of associated `CXConfig` objects.
        :returns: `List[CXConfig]`
        """
        return self.__configs

    def render_to_dict(self) -> dict:
        """
        Provides a `dict` representation of the configuration values.
        :returns: `dict`
            A `dict` representing the configuration values arranged as a map
            of keys and values.

            Given:
            ```python
            configs = CXConfigs()
            configs \
                .set_param("1", "rgb(3, 172, 198)") \
                .set_param("2", 2) \
                .set_param("3", True)
            ```

            Then `render_to_dict()` results in:
            ```python
            {
                "1": "rgb(3, 172, 198)",
                "2": 2,
                "3": True,
            }
            ```
        """
        merged_configs = CXConfigs.merge_configs(self.configs)

        # Remove illegal options that cause crashes
        if merged_configs.get("isHistogram") is not None:
            del merged_configs["isHistogram"]

        return merged_configs

    def render_to_list(self, **kwargs) -> list:
        """
        Provides a `list` representation of the configuration values.
        :returns: `list`
            A `list` representing the configuration values arranged as a map
            of keys and values.

            Given:
            ```python
            configs = CXConfigs()
            configs \
                .set_param("1", "rgb(3, 172, 198)") \
                .set_param("2", 2) \
                .set_param("3", True)
            ```

            Then `render_to_list()` results in:
            ```python
            [
                ["1", "rgb(3, 172, 198)"],
                ["2": 2],
                ["3": True],
            ]
            ```
        """
        configs = self.render_to_dict()
        return [
            [str(key), configs[key]]  # function name  # list of parameter values
            for key in configs.keys()
        ]

    @classmethod
    def merge_configs(cls, configs: List[CXConfig]) -> dict:
        """
        Given a list of CXConfig objects, a dictionary of unique attributes is
        generated and provided.
        :returns: `dict`
            A dict of zero or more keys representing the CXConfigs.
        """
        unique_configs = list()
        if not configs is None:
            for config in configs:
                if not config in unique_configs:
                    unique_configs.append(config)

        dict_configs = dict()
        for config in unique_configs:
            dict_configs = {**dict_configs, **(config.render())}

        return dict_configs

    def __copy__(self) -> "CXConfigs":
        """
        *copy* constructor.  Returns the `CXConfig` objects within a new
        `CXConfigs` object.
        """
        return CXConfigs(*self.configs)

    def __deepcopy__(self, memo) -> "CXConfigs":
        """
        *deepcopy* constructor.  Returns a deepcopy of the `CXConfig` objects
         within a new `CXConfigs` object.
        """
        return CXConfigs(*([deepcopy(config) for config in self.configs]))

    def __lt__(self, other: "CXConfigs") -> bool:
        """
        *less than* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXConfigs` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXConfigs` object then False
            <li> If `other` is a `CXConfigs` object then True of all `CXConfig`
                objects are also less than the events tracked by `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXConfigs):
            return False

        else:
            if (len(self.configs) + len(other.configs)) == 0:
                return False

            if len(self.configs) == len(other.configs):
                for config in self.configs:
                    for oconfig in other.configs:
                        if not config < oconfig:
                            return False
                return True

            else:
                return len(self.configs) < len(other.configs)

    def __eq__(self, other: "CXConfigs") -> bool:
        """
        *equals* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXConfigs` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXConfigs` object then False
            <li> If `other` is a `CXConfigs` object then True of all `CXConfig`
                objects are also equal to the events tracked by `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXConfigs):
            return False

        else:
            if len(self.configs) == len(other.configs):
                for config in self.configs:
                    alt_config = other.get_param(config.label)
                    if alt_config is None:
                        return False
                    if alt_config.value != config.value:
                        return False

                return True

            else:
                return len(self.configs) == len(other.configs)

    def __str__(self) -> str:
        """
        *str* function.  Converts the `CXConfigs` object into a JSON
         representation.
        :returns" `str`
            JSON form of the collection.
        """
        return json.dumps(self.render_to_dict())

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the `CXConfigs` object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        config_rep_list = ", ".join([repr(config) for config in self.configs])
        rep_candidate = f"CXConfigs(" f"{config_rep_list}" f")"
        return rep_candidate



"""
This module provides an integration solution for the CanvasXpress class and the base dash component for a feature-
rich integration with Plotly's Dash framework.
"""

import json
from typing import List, Union, Any

from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.base import CXRenderFactory
from cxdash import CXDashElement


class CXElementFactory(CXRenderFactory):
    """
    CXDashElementFactory converts CanvasXpress objects into PlotlyDash elements that will
    render CanvasXpress charts into the Dash UI.
    """

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXDashRenderFactory` object.
        :param cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        super().__init__(*cx)

    @classmethod
    def render(cls, cx: CanvasXpress) -> CXDashElement:
        """
        Converts the provided CanvasXpress object into a CXDashElement object that is
        primed by the configuration represented by the CanvasXpress object.
        :param cx: `CanvasXpress` The CanvasXpress object to render.  An exception is
            raised if cx is `None`.
        :returns: `CXDashElement` A CXDashElement with the configuration as represented
            by cx.
        """
        element_parts = cx.prepare_html_element_parts()
        dash_element = CXDashElement(
            id=element_parts["renderTo"],
            data=(
                element_parts["data"]
                if isinstance(element_parts["data"], str)
                else json.dumps(element_parts["data"])
            ),
            config=json.dumps(
                {**element_parts["config"], **element_parts["otherParams"]}
            ),
            events=element_parts["events"],
            after_render=json.dumps(element_parts["afterRender"]),
            js_url=cx.js_library_url(),
            css_url=cx.css_library_url(),
            width=str(element_parts["width"]),
            height=str(element_parts["height"]),
        )

        return dash_element

    def render_all(self, **kwargs: Any) -> List[CXDashElement]:
        """
        Provides a list of objects that can be used by the target domain or container
        to create CanvasXpress illustrations or instantiations.
        Not implemented.
        :param kwargs: `Any`
            Parameters specific to implementations are supported.  The essential
            create_element call should work with no extra parameters, and with
            parameters that do not apply to the implementation.
        """
        if isinstance(self.canvas, CanvasXpress):
            return [CXElementFactory.render(self.canvas)]

        else:
            return [CXElementFactory.render(cx) for cx in self.canvas]



import csv
import json
from typing import Union

from pandas import DataFrame

from canvasxpress.data.base import CXData


class CXTextData(CXData):
    """
    `CXTextData` is a `CXData` class that provides plain-text data directly to
    the CanvasXpress for Javascript object.  In this manner, the Python tier
    makes no assumptions about the data content and permits the Javascript tier
    to address any required adjustments in order to properly display the data
    within a chart.  If the data is erroneously formatted then the only
    feedback will be at the Javascript tier.
    """

    __raw_text = ""
    """
    `__raw_text` tracks a block of text to be passed directly to the 
    CanvasXpress for Javascript constructor.
    """

    @property
    def text(self) -> str:
        """
        Returns the raw text form of the data.
        :returns: `str`
            The text to be provided to CanvasXpress.
        """
        return self.__raw_text

    @text.setter
    def text(self, value: str) -> None:
        """
        Sets the text to be provided to CanvasXpress.
        :param value: `str`
            The text to provide as-is to CanvasXpress.  `None` will be
            converted to an empty `str`.  Values of type other than `str`
            will be converted using `str()`.
        """
        if value is None:
            self.__raw_text = ""

        elif isinstance(value, str):
            self.__raw_text = value

        else:
            self.__raw_text = str(value)

    @property
    def data(self) -> dict:
        """
        A property accessor for the data managed by the object.  Regardless of
        the input data the returned data structure will be a dict-type for use
        with CanvasXpress.
        :returns: `dict`
            A dictionary representing a data map suitable for use with a chart.
        """
        return self.get_raw_dict_form()

    def get_raw_dict_form(self) -> dict:
        """
        Provides a simple dict perspective of the data with no metadata or other
        contextual transformations performed.  For example, if the data is
        natively in `dict` form then it would be passed-through with no
        modification or enhancement.
        :returns: `dict`
            The `dict` perspective of the data with as little modification or
            interpretation as is reasonable.
        """
        try:
            # Check the data as a JSON object.  If the JSON object equates to
            # a dict, list, or str then pass the Python form along as it will be
            # converted back into a string as part of the HTML render.  For
            # anything else treat the content as a standard string to be
            # passed along.

            candidate = {"raw": json.loads(self.text)}
            if isinstance(candidate["raw"], (dict, list, str)):
                return {"raw": json.loads(self.text)}
            else:
                return {"raw": self.text}

        except Exception as e:
            return {"raw": self.text}

    def render_to_dict(self, **kwargs) -> dict:
        """
        Converts the object into a dict representation.
        :returns: `dict`
            A dictionary representation of the object, such as what might be
            needed for a JSON export.
        """
        return self.get_raw_dict_form()

    def __init__(self, data: Union[object, None] = None) -> None:
        """
        Initializes the CXData object with data.
        :param data: `Union[object, None]`
            Given an object or no data prepares a new CXData instance ready for
            use by a `CanvasXpress` object.
        """
        self.text = data


class CXDataframeData(CXTextData):
    """
    `CXDataframeData` is a `CXTextData` class that provides plain-text data directly to
    the CanvasXpress for Javascript object by converting a Pandas DataFrame into a text matrix.
    """

    __data: DataFrame = DataFrame()
    """
    The data managed by an object of this class.
    """

    __delimiter: str = ","
    """
    The delimiter to use when convert to a CSV or TSV.
    """

    @property
    def delimiter(self) -> str:
        """
        Provides the delimiter (separator) used when creating the text matrix.
        """
        return self.__delimiter

    @delimiter.setter
    def delimiter(self, value) -> None:
        """
        Sets the delimiter to the desired value.  Must be CSV or TSV compliant.
        """
        if not isinstance(value, str):
            raise TypeError(
                "The delimiter must be a single character str bearing a comma or tab."
            )

        if value not in [",", "\t"]:
            raise ValueError(
                "The delimiter must be a single character str bearing a comma or tab."
            )

        else:
            self.__delimiter = value

    @property
    def dataframe(self) -> DataFrame:
        """
        Provides the data managed by the object.
        :returns: `DataFrame` The managed data.
        """
        return self.__data

    @dataframe.setter
    def dataframe(self, value: Union[DataFrame, None] = None) -> None:
        """
        Sets the dataframe managed by the object.
        :param value: `Union[DataFrame, None]`
            `None` results in an empty `DataFrame`.  A deepcopy will be made of
            `DataFrame` values.
        """

        if value is not None and not isinstance(value, DataFrame):
            raise TypeError("The value of dataframe must be a valid DataFrame or None.")

        self.dataframe = value

        if self.dataframe is None:
            self.text = None

        else:
            self.text = self.dataframe.to_csv(
                index=False,
                sep=self.delimiter,
                quoting=csv.QUOTE_NONNUMERIC,
            )

    @property
    def text(self) -> str:
        return super().text

    @text.setter
    def text(self, value: str) -> None:
        """
        Sets the text to be provided to CanvasXpress.
        :param value: `str`
            The text to provide as-is to CanvasXpress.  `None` will be
            converted to an empty `str`.  Values of type other than `str`
            will be converted using `str()`.
        """
        if value is None:
            self.__raw_text = ""

        elif isinstance(value, str):
            self.__raw_text = value

        else:
            self.__raw_text = str(value)

    def __init__(
        self,
        data: Union[DataFrame, None] = None,
        delimiter: str = ",",
    ) -> None:
        """
        Initializes the CXDataFrameData object with data.
        :param data: `Union[DataFrame, None]`
            Given an DataFrame or no data prepares a new CXDataFrameData instance ready for
            use by a `CanvasXpress` object.
        :param delimiter: `str`
            A single character `str` with a comma or tab as the value.  The default is a comma.
        """
        super().__init__(None)
        self.delimiter = delimiter
        self.dataframe = data

    def __copy__(self) -> "CXDataframeData":
        """
        *copy constructor* that returns a copy of the CXDataframeData object.
        :returns: `CXDataframeData`
            A copy of the wrapping object.
        """
        return self.__class__(self.dataframe)

    def __deepcopy__(self, memo) -> "CXDataframeData":
        """
        *deepcopy constructor* that returns a copy of the CXDataframeData object.
        :returns: `CXDataframeData` A copy of the wrapping object and deepcopy of
            the tracked data.
        """
        return self.__class__(self.data)

    def __lt__(self, other: "CXDataframeData") -> bool:
        """
        *less than* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXDataframeData` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXDataframeData` object then False
            <li> If `other` is a `CXDataframeData` object then True of all
                `CXDataframeData` aspects are also less than the data tracked by
                `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXDataframeData):
            return False

        else:
            self_c = self.dataframe.columns.unique()
            other_c = other.dataframe.columns.unique()

            if len(self_c) < len(other_c):
                return True

            elif len(self_c) > len(other_c):
                return False

            else:
                for i in [s for s in self_c if s not in other_c]:
                    if any([i < o for o in other_c]):
                        return True

                return self.dataframe.lt(other.dataframe).all(axis=None)

    def __eq__(self, other: "CXDataframeData") -> bool:
        """
        *equals* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXDataframeData` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXDataframeData` object then False
            <li> If `other` is a `CXDataframeData` object then True of all
                `CXDataframeData` aspects are also less than the data tracked by
                `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXDataframeData):
            return False

        else:
            self_c = self.dataframe.columns.unique()
            other_c = other.dataframe.columns.unique()

            if len(self_c) != len(other_c):
                return False

            if any([s not in other_c for s in self_c]):
                return False

            return self.dataframe.eq(other.dataframe).all(axis=None)

    def __str__(self) -> str:
        """
        *str* function.  Converts the CXDataframeData object into a JSON
        representation.
        :returns" `str` JSON form of the `CXDataframeData`.
        """
        return json.dumps(self.render_to_dict())

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXDataframeData object into a pickle
         string that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        candidate = (
            f"CXDataframeData("
            f"data=pandas.read_csv("
            f'StringIO("""{self.text}"""),'
            f"index_col=0))"
        )
        candidate = candidate.replace("Infinity", "float('inf')")
        candidate = candidate.replace("NaN", "float('nan')")

        return candidate



def render_from_template(template: str, data: dict) -> str:
    """
    Updates the template text with the provided data.
    :param template: `str` The name of the template file
    :param data: The `dict` of str values with which to update the template text
    :returns The adjusted template text
    """
    template_text = template

    for key in data.keys():
        template_text = template_text.replace(f"@{str(key)}@", str(data[key]))

    return template_text



from typing import Any, Union

from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.environment import (
    get_target_context,
    CONTEXT_RSTUDIO,
    CONTEXT_SHINY,
    CONTEXT_DASH,
    CONTEXT_JUPYTER,
    CONTEXT_STREAMLIT,
    CONTEXT_BROWSER,
)
from canvasxpress.render.image import CXImage
from canvasxpress.render.json import CXJSON
from canvasxpress.render.popup import CXBrowserPopup

# Track the runtime context
_g_contexts_imported: list = []
_g_context = get_target_context()

# Pre-load Jupyter JS and CSS.
if _g_context == CONTEXT_JUPYTER:
    if not _g_context in _g_contexts_imported:
        from canvasxpress.render.jupyter import CXNoteBook

        CXNoteBook.display_canvasxpress_header()


def convert_from_reproducible_json(json: str) -> Union[None, CanvasXpress]:
    """
    Accepts a str with a reproducible JSON and returns a CanvasXpress object.
    """
    return CanvasXpress.from_reproducible_json(json)


def convert_to_reproducible_json(canvas: CanvasXpress) -> str:
    """
    Converts the CanvasXpress object into a reproducible JSON string.
    """
    return CXJSON.render_to_json(canvas)


def convert_to_image(canvas: CanvasXpress, type: str = "png") -> Union[None, bytes]:
    """
    Converts the CanvasXpress object to an image of the specified type.
    """
    converter = CXImage(canvas)
    candidates = converter.render(format=type)
    for conversion in candidates:
        if conversion.get("image", {}).get("format") == type:
            return conversion.get("image", {}).get("binary")


def show_in_browser(canvas: CanvasXpress) -> None:
    """
    Opens a browser and displays the canvas.
    """
    plotter = CXBrowserPopup(canvas)
    plotter.render()


def graph(canvas: CanvasXpress) -> Any:
    """
    Displays the CanvasXpress object as a visualized chart in a manner appropriate to the running context.

    To override an assumed context the ENV variable `CANVASXPRESS_TARGET_CONTEXT` can be set to one of the following
    values, in which case `show` will attempt to illustrate the chart of the set target.

    - rstudio
    - shiny
    - jupyter
    - dash
    - streamlit
    - browser

    :returns: An `object` or `None` depending on the target context.  In the case of `browser` a popup browser will
        be launched.`
    """
    if _g_context == CONTEXT_RSTUDIO:
        if not _g_context in _g_contexts_imported:
            from canvasxpress.render.shiny import CXShinyWidget

        plotter = CXShinyWidget(canvas)
        plotter._repr_rstudio_viewer_()

    elif _g_context == CONTEXT_SHINY:
        if not _g_context in _g_contexts_imported:
            from canvasxpress.render.shiny import CXShinyWidget

        plotter = CXShinyWidget(canvas)
        return plotter

    elif _g_context == CONTEXT_DASH:
        if not _g_context in _g_contexts_imported:
            from canvasxpress.render.dash import CXElementFactory

        plotter = CXElementFactory()
        return plotter.render(canvas)

    elif _g_context == CONTEXT_JUPYTER:
        if not _g_context in _g_contexts_imported:
            from canvasxpress.render.jupyter import CXNoteBook

        plotter = CXNoteBook(canvas)
        return plotter.render()

    elif _g_context == CONTEXT_STREAMLIT:
        if not _g_context in _g_contexts_imported:
            from canvasxpress.render import streamlit

        streamlit.plot(canvas)

    elif _g_context == CONTEXT_BROWSER:
        plotter = CXBrowserPopup(canvas)
        plotter.render()

    else:
        return """
        show() cannot identify the target context.  Either explicitly use a module from canvasxpress.render to 
        illustrate the CanvasXpress object, which requires the proper canvasxpress package option to be installed 
        (for example, pip install canvasxpress[shiny]) or set the environment variable CANVASXPRESS_TARGET_CONTEXT 
        to be one of rstudio, shiny, jupyter, streamlit, dash, or browser.
        """



import uuid
from pathlib import Path
from typing import Any, Union, List

import htmlmin
from IPython.display import HTML
from bs4 import BeautifulSoup

from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.base import CXRenderable

_cx_intermixed_header = """
<html>
    <meta charset="UTF-8">
    <link 
        href='@css_url@' 
        rel='stylesheet' 
        type='text/css'
    />
    <script 
        src='@js_url@' 
        type='text/javascript'>
    </script>
</html>
"""

_cx_js_intermixed_template = """
<script type="text/javascript">
    @code@
</script>
"""

_cx_html_intermixed_template = """
<html>
    @canvasxpress_license@
    @canvases@
    @js_functions@
</html>
"""


class CXHtml(CXRenderable):
    """
    CXHtml is a `CXRenderable` that renders `CanvasXpress` objects into
    HTML Snippets suitable for use in RStudio
    """

    def __init__(self, *cx: Union[List[CanvasXpress], CanvasXpress, None]):
        """
        Initializes a new `CXHtml` object.
        :praram cx: `Union[List[CanvasXpress], CanvasXpress, None], ...`
            The `CanvasXpress` object(s) to be tracked.  See the `canvas`
            property, except that on initialization cx can be `None`.
            Multiple CanvasXpress objects are supported provided that
            they have distinct `render_to` targets.
        """
        super().__init__(*cx)

    def display_canvasxpress_header(self):
        css_url = CanvasXpress.css_library_url()
        js_url = CanvasXpress.js_library_url()

        header_html_text = _cx_intermixed_header.replace("@css_url@", css_url).replace(
            "@js_url@", js_url
        )

        return header_html_text

    def display_debug_code(self, code: str):
        minified_code = htmlmin.Minifier().minify(code)
        pretty_code = BeautifulSoup(minified_code, "html.parser").prettify()
        return pretty_code

    def get_chart_display_code(self, columns: int) -> str:
        render_targets = list()

        if self.canvas is None:
            pass

        elif isinstance(self.canvas, CanvasXpress):
            render_targets.append(self.canvas)

        else:
            render_targets.extend(self.canvas)

        used_render_targets = list()
        for target in render_targets:
            original_render_target = target.render_to
            if original_render_target in used_render_targets:
                target.render_to = (
                    original_render_target + "_" + str(uuid.uuid4()).replace("-", "_")
                )

            used_render_targets.append(target.render_to)

        render_targets.reverse()

        html_parts = [target.render_to_html_parts() for target in render_targets]

        canvases = [part["cx_canvas"] for part in html_parts]
        if len(canvases) < columns:
            columns = len(canvases)

        functions = [part["cx_js"] for part in html_parts]

        cx_license = ""
        for part in html_parts:
            if part.get("cx_license"):
                cx_license = part["cx_license"]
                break

        canvas_table = "<table>"
        column_index = 0
        for chart_index in range(len(canvases)):
            if column_index == 0:
                canvas_table += "<tr>"

            canvas_table += '<td style="padding: 5px"><div>'
            canvas_table += canvases[chart_index]
            canvas_table += "</div></td>"

            column_index += 1
            if column_index == columns:
                column_index = 0
                canvas_table += "</tr>"

        canvas_table += "</table>"

        js_functions = "\n".join(
            [_cx_js_intermixed_template.replace("@code@", fx) for fx in functions]
        )
        html_text = (
            _cx_html_intermixed_template.replace("@canvases@", canvas_table)
            .replace("@canvasxpress_license@", cx_license)
            .replace("@js_functions@", js_functions)
        )

        return html_text

    def display_charts(self, code: str, output_file: str):
        try:
            if output_file is not None:
                file_path_candidate = str(output_file)
                file_path = Path(file_path_candidate)
                if file_path.is_dir():
                    file_path = file_path.joinpath(f"cx_{str(uuid.uuid4())}.html")

                with open(str(file_path), "w") as render_file:
                    render_file.write(code)

        except Exception as e:
            raise RuntimeError(f"Cannot create output file: {e}")

        return code

    def render(self, **kwargs: Any):
        """
        Renders the associated CanvasXpress object appropriate for display in
        a RStudio environment.  Charts cannot have the same name, so render_to
        will be updated with a uuid for each conflicting chart.
        :param kwargs: `Any`
            * Supports `columns` for any positive `int` of `1` or greater, with a
              default value of `1`.  Values less that `1` are ignored.  `columns`
              indicates how many charts should be rendered horizontally if more
              than one chart is being tracked.
            * Supports `output_file` as a string for a path at which the output
              should be saved.  If a file exists at the specified path then
              it will be overwritten.
            * Supports `debug` for displaying the output source.  True indicates
              that the HTML code shall be displayed prior to the parsed output.
              Default is False.
        """
        debug_output_arg = kwargs.get("debug")
        debug_output = bool(debug_output_arg) if debug_output_arg is not None else False

        columns_arg = int(kwargs.get("columns", 1))
        columns = columns_arg if columns_arg > 0 else 1

        output_file_arg = kwargs.get("output_file")
        output_file = (
            output_file_arg
            if output_file_arg is not None and isinstance(output_file_arg, str)
            else None
        )

        code = self.get_chart_display_code(columns)

        try:
            header = self.display_canvasxpress_header()
            body = self.display_charts(code, output_file)
            if debug_output:
                self.display_debug_code(code)

            return HTML(header + body)

        except Exception as e:
            raise RuntimeError(f"Cannot create output cell: {e}")



import html
import json
from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from functools import total_ordering
from typing import Union, Any

from deepdiff import DeepDiff
from deprecated import deprecated


@total_ordering
class CXConfig(ABC):
    """
    CXConfig provides the means by which CanvasXpress objects can be configured for
    customized rendering and interaction.
    """

    __label: str = ""
    """
    The configuration object's label.
    """

    @property
    def label(self) -> str:
        """
        Provides the label for the configuration.
        :returns: `str`
        """
        return self.__label

    @property
    @abstractmethod
    def value(self) -> Any:
        """
        Provides the value for the configuration.  Must be implemented by
        concrete classes.
        :returns: `Any`
            The relevant type of value.
        """
        pass

    @value.setter
    @abstractmethod
    def value(self, value: Any) -> None:
        """
        Sets the value of the configuration.  Must be implemented by concrete
        classes.
        :param value: `Any`
            The value to be accepted.  Will be more specific with concrete
            implementations, such as `str` for string configurations.
        """
        pass

    def render(self) -> dict:
        """
        Renders the value in a form suitable for use in preparing Javascript.
        Typically, this will be the native `value`.

        :returns: `dict`
            A version of the `value` most appropriate for use in prepating the
            Javascript rendering.
        """
        return {self.label: self.value}

    def __init__(self, label: str, value: Any):
        """
        Initializes a new CXConfig object with a label and value.
        :param label: `str`
            The label for the configuration.
        :param value: `Any`
            The value for the configuration.  See the `value` property for the
            concrete implementation for allowed types.
        """
        if label is None:
            raise ValueError("label cannot be None")
        self.__label = label

    def __copy__(self) -> "CXConfig":
        """
        *copy constructor* that provides a new CXConfig of the same type with
        the data referenced.
        :returns: `CXConfig` of the proper type
        """
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        """
        *deepcopy constructor* that provides a new CXConfig of the same type with
        the a deepcopy of the data.
        :returns: `CXConfig` of the proper type
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result

    def __hash__(self) -> int:
        """
        Provides a hash proxy for the object as converted into its `repr` form.
        :returns: `int`
        """
        return hash(repr(self))

    def __lt__(self, other: "CXConfig") -> bool:
        """
        *less than* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXConfig` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXConfig` object then `False`
            <li> If `other` is a `CXConfig` object then True if label and value
                of `other` are less than that of `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, self.__class__):
            return False

        if self.label < other.label:
            return True

        elif self.label == other.label:
            if self.value < other.value:
                return True

            else:
                return False

    def __eq__(self, other: "CXConfig") -> bool:
        """
        *equals* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXConfig` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXConfig` object then `False`
            <li> If `other` is a `CXConfig` object then True if label and value
                of `other` are equal to that of `self`.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, self.__class__):
            return False

        return (self.label == other.label) and (self.value == other.value)

    def __str__(self) -> str:
        """
        *str* function.  Converts the object into a JSON string.
        """
        return json.dumps(
            {
                "label": self.label,
                "value": self.value,
            }
        )

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXConfig object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        return (
            f"{str(self.__class__).split('.')[-1][:-2]}("
            f" label='{self.label}',"
            f" value={json.dumps(self.value)}"
            f")"
        )


class CXString(CXConfig):
    """
    A `CXConfig` object that manages `str` values.
    """

    __value: str = ""
    """
    The managed value.
    """

    @property
    def value(self) -> str:
        """
        Provides the value for the configuration.
        :returns: `str`
        """
        return self.__value

    @value.setter
    def value(self, value: Union[object, str]) -> None:
        """
        Sets the value of the configuration.
        :param value: `str`
            If `None` then an empty `str` will be used.
        """
        if value is None:
            self.__value = ""
        else:
            self.__value = str(value)

    def __init__(self, label: str, value: str):
        """
        Initializes the configuration with a `str` value.
        """
        super().__init__(label, value)
        self.value = value


class CXNone(CXConfig):
    """
    A `CXConfig` object that manages `None` values.
    """

    __value: Any = None
    """
    The managed value.
    """

    @property
    def value(self) -> None:
        """
        Provides the value for the configuration.
        :returns: `str`
        """
        return self.__value

    @value.setter
    def value(self, value=None) -> None:
        """
        Sets the value of the configuration.
        :param value: `None`
            Only None can be used.
        """
        if value is not None:
            raise ValueError("CXNone only accepts None values")

        self.__value = None

    def __init__(self, label: str, value=None):
        """
        Initializes the configuration with a None value.
        """
        super().__init__(label, value)
        self.value = value


class CXBool(CXConfig):
    """
    A `CXConfig` object that manages `bool` values.
    """

    __value: bool = False
    """
    The managed value.
    """

    @property
    def value(self) -> bool:
        """
        Provides the value for the configuration.
        :returns: `bool`
        """
        return self.__value

    @value.setter
    def value(self, value: Union[object, bool]) -> None:
        """
        Sets the value of the configuration.
        :param value: `bool`
            If `None` then `False` will be used.
        """
        if value is None:
            self.__value = False
        else:
            self.__value = bool(value)

    def __init__(self, label: str, value: bool):
        """
        Initializes the configuration with a `bool` value.
        """
        super().__init__(label, value)
        self.value = value

    def __str__(self) -> str:
        """
        *str* function.  Converts the object into a Javascript statement.
        """
        return str(
            {
                "label": self.label,
                "value": self.value,
            }
        )

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXBool object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        return (
            f"{str(self.__class__).split('.')[-1][:-2]}("
            f" label='{self.label}',"
            f" value={str(self.value)}"
            f")"
        )


class CXFloat(CXConfig):
    """
    A `CXConfig` object that manages `float` values.
    """

    __value: float = 0.0
    """
    The managed value.
    """

    @property
    def value(self) -> float:
        """
        Provides the value for the configuration.
        :returns: `float`
        """
        return self.__value

    @value.setter
    def value(self, value: Union[object, float]) -> None:
        """
        Sets the value of the configuration.
        :param value: `float`
            If `None` then `float(0.0)` will be used.
        """
        if value is None:
            self.__value = 0.0
        else:
            self.__value = float(value)

    def __init__(self, label: str, value: float):
        """
        Initializes the configuration with a `float` value.
        """
        super().__init__(label, value)
        self.__value = 0.0

        self.value = value


class CXInt(CXConfig):
    """
    A `CXConfig` object that manages `int` values.
    """

    __value: int = 0
    """
    The managed value.
    """

    @property
    def value(self) -> int:
        """
        Provides the value for the configuration.
        :returns: `int`
        """
        return self.__value

    @value.setter
    def value(self, value: Union[object, int]) -> None:
        """
        Sets the value of the configuration.
        :param value: `int`
            If `None` then `int(0)` will be used.
        """
        if value is None:
            self.__value = 0
        else:
            self.__value = int(value)

    def __init__(self, label: str, value: int):
        """
        Initializes the configuration with an `int` value.
        """
        super().__init__(label, value)
        self.__value = 0

        self.value = value


class CXDict(CXConfig):
    """
    A `CXConfig` object that manages `dict` values.
    """

    __value: dict = dict()
    """
    The managed value.
    """

    @property
    def value(self) -> dict:
        """
        Provides the value for the configuration.
        :returns: `dict`
        """
        return self.__value

    @value.setter
    def value(self, value: Union[dict, str, None]) -> None:
        """
        Sets the value of the configuration.
        :param value: `dict`
            If `None` then `dict()` will be used.
        """
        if value is None:
            self.__value = dict()

        elif isinstance(value, CXDict):
            self.__value = deepcopy(value.value)

        elif isinstance(value, str):
            candidate = json.loads(html.unescape(value).replace("&nl;", '"<br>"'))
            self.__value = candidate

        else:  # remove html entities from string values
            clean_value = {}
            for k, v in value.items():
                clean_value[k] = v
                if isinstance(v, str):
                    clean_value[k] = html.unescape(v).replace("&nl;", '"<br>"')
            self.__value = deepcopy(clean_value)

    def __init__(self, label: str, value: Union[dict, str, None]) -> None:
        """
        Initializes the CXData object with data.  Only dict or compatible data
        types are accepted.
        """
        super().__init__(label, value)
        self.value = value

    def __lt__(self, other: "CXDict") -> bool:
        """
        *less than* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXDict` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXDict` object then `False`
            <li> If `other` is a `CXDict` object then True if the label and
                value parts are less than that of self.
            </ul>
        """
        if other is None:
            return False

        if not isinstance(other, CXDict):
            return False

        if self.label < other.label:
            return True

        if self.label > other.label:
            return False

        else:
            delta: dict = DeepDiff(self.value, other.value, ignore_order=True)
            other_added: int = len(delta.get("dictionary_item_added", []))
            other_removed: int = len(delta.get("dictionary_item_removed", []))

            if (other_added - other_removed) == 0:
                for skey in self.value.keys():
                    if skey not in other.value.keys():
                        for okey in other.value.keys():
                            if skey < okey:
                                return True

                    elif self.value[skey] < other.value[skey]:
                        return True

                return False

            else:
                return (other_added - other_removed) > 0

    def __eq__(self, other: "CXDict") -> bool:
        """
        *equals* comparison.  Also see `@total_ordering` in `functools`.
        :param other:
            `CXDict` The object to compare.
        :returns: `bool`
            <ul>
            <li> If `other` is `None` then `False`
            <li> If `other` is not a `CXDict` object then `False`
            <li> If `other` is a `CXDict` object then True if the label and
                value parts are equal to that of self.
            </ul>
        """
        if other is None:
            return False

        elif not isinstance(other, CXDict):
            return False

        elif self.label < other.label:
            return False

        elif self.label > other.label:
            return False

        else:
            delta: dict = DeepDiff(self.value, other.value, ignore_order=True)
            other_added: int = len(delta.get("dictionary_item_added", []))
            other_removed: int = len(delta.get("dictionary_item_removed", []))

            if (other_added - other_removed) == 0:
                for skey in self.value.keys():
                    if skey not in other.value.keys():
                        return False

                    elif self.value[skey] != other.value[skey]:
                        return False

                return True

            else:
                return False

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXDict object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        return f"CXDict(label='{self.label}', value={json.dumps(self.value)})"


class CXList(CXConfig):
    """
    A `CXConfig` object that manages `list` values.
    """

    __value: list = list()
    """
    The managed value.
    """

    @property
    def value(self) -> list:
        """
        Provides the value for the configuration.
        :returns: `list`
        """
        return self.__value

    @value.setter
    def value(self, value: Union[object, list]) -> None:
        """
        Sets the value of the configuration.
        :param value: `list`
            If `None` then `list()` will be used.
        """
        if value is None:
            self.__value = list()
        else:
            self.__value = list(value)

    def __init__(self, label: str, value: list):
        """
        Initializes the configuration with a `list` value.
        """
        super().__init__(label, value)
        self.__value = list()

        self.value = value


class CXGraphWeight(CXConfig):
    """
    A `CXConfig` object that manages and normalizes `CXGraphWeight list` values.
    """

    __value: list = list()
    """
    The managed value.
    """

    def is_graph_weight_list(label: str, value: list):
        """
        A static method that evaluates a given list and label to check
        if it represents  graph weight list.
        :param label: 'str'
            A string represents the value label i.e "ringGraphWeight".
            It must contain "weight" word
        :param value: `list`
            A list to evaluate.  It must be:
              - non-empty list
              - all its values are numeric
              - summation of its values is 100
        :returns: `bool`
            True if all conditions are valid
        """
        valid_label = False
        if label is not None:
            valid_label = "weight" in label.lower()
        valid_list = True

        if valid_label:
            if value is None:
                valid_list = False
            elif not isinstance(value, list):
                valid_list = False
            elif not all(isinstance(x, (int, float)) for x in value):
                valid_list = False
            elif sum(value) != 100 and sum(value) != 1:
                valid_list = False

        return valid_label and valid_list

    @property
    def value(self) -> list:
        """
        Provides the value for the configuration.
        :returns: `list`
        """
        return self.__value

    @value.setter
    def value(self, value: Union[object, list]) -> None:
        """
        Sets the value of the configuration.
        :param value: `list`
            If `None` then `list()` will be used.
        """
        # return empty value if the weight list or its sum is invalid
        final_value = list()

        # normalize graph weight only if it has non normalized and valid weight list
        if value is not None and all(isinstance(x, (int, float)) for x in value):
            if sum(value) == 100:
                final_value = [x / 100 for x in value]
            elif sum(value) == 1:
                final_value = value

        self.__value = final_value

    def __init__(self, label: str, value: list):
        """
        Initializes the configuration with a `list` value.
        """
        super().__init__(label, value)
        self.__value = list()

        self.value = value


@deprecated(action="ignore")
class CXRGBAColor(CXDict):
    """
    A `CXConfig` object that manages `str` Javascript rgba() values.
    """

    @staticmethod
    def is_color_str(value: str):
        """
        A static method that evaluates a given string to see if it represents a
        Javascript rgba() statement.
        :param value: `str`
            A string to evaluate.  A valid Javascript value has the form
            `rgba(r, g, b, a)` where RGB values are `int` from 0-255 and A is a
            `float` from 0.0 to 1.0.
        :returns: `bool`
            True if the string represents a Javascript rgba() statement.
        """
        if isinstance(value, str):
            if not value.startswith("rgba"):
                return False

            try:
                components = value.split(",")
                if len(components) != 4:
                    return False

                r = components[0].strip().split("rgba(")[1]
                g = components[1].strip()
                b = components[2].strip()
                a = components[3].strip().split(")")[0]

                for x in [int(r), int(g), int(b)]:
                    if (x < 0) or (x > 255):
                        return False
                for x in [float(a)]:
                    if (x < 0) or (x > 1):
                        return False

                return True

            except:
                return False

        else:
            return False

    @staticmethod
    def is_color_list(value: list):
        """
        A static method that evaluates a given list to see if it represents a
        Javascript rgba() statement.
        :param value: `list`
            A list to evaluate.  A valid Javascript value has the form
            `rgba(r, g, b, a)` where RGB values are `int` from 0-255 and A is a
            `float` from 0.0 to 1.0.
        :returns: `bool`
            True if the list represents a Javascript rgba() statement.
        """
        if isinstance(value, list):
            try:
                list_len = len(value)
                all_rgb_elements_int = all(isinstance(x, int) for x in value[:3])
                alpha_element_num = type(value[3]) in [int, float]
                if (
                    (not list_len == 4)
                    or (not all_rgb_elements_int)
                    or (not alpha_element_num)
                ):
                    return False

                for x in value[:3]:
                    if (x < 0) or (x > 255):
                        return False
                if (value[3] < 0) or (value[3] > 1):
                    return False

                return True

            except:
                return False

        else:
            return False

    @staticmethod
    def is_color_dict(value: dict):
        """
        A static method that evaluates a given dict to see if it represents a
        Javascript rgba() statement.
        :param value: `dict`
            A dict to evaluate.  A valid Javascript value has the form
            `rgba(r, g, b, a)` where RGB values are `int` from 0-255 and A is a
            `float` from 0.0 to 1.0.  For the dict to be valid its keys must be
            lower case r, g, b, and a characters.
        :returns: `bool`
            True if the dict represents a Javascript rgba() statement.
        """
        if isinstance(value, dict):
            list_len = len(value.keys())
            all_rgba_elements = all(x in ["r", "g", "b", "a"] for x in value.keys())
            all_rgb_elements_int = all(
                isinstance(value[x], int) for x in value.keys() if x in ["r", "g", "b"]
            )
            alpha_element_num = type(value.get("a")) in [int, float]
            keys_rgba = all(
                [j in ["r", "g", "b", "a"] for j in [k for k in value.keys()]]
            )
            if (
                (not list_len == 4)
                or (not all_rgb_elements_int)
                or (not keys_rgba)
                or (not alpha_element_num)
                or (not all_rgba_elements)
            ):
                return False

            for key in value.keys():
                if key in ["r", "g", "b"]:
                    if (value[key] < 0) or (value[key] > 255):
                        return False
                if key in ["a"]:
                    if (value[key] < 0) or (value[key] > 1):
                        return False

            return True

        else:
            return False

    @CXDict.value.setter
    def value(self, value: Union["CXRGBAColor", dict, list, str]) -> None:
        """
        Sets the RGBA value from an existing `CXRGBAColor` object, or a `dict`,
        `list`, or `string` following the Javascript `rgba()` format.
        :param value: `Union['CXRGBAColor', dict, list, str]`
            The value to be accepted.  See the `is_color_*()` methods for
            acceptable formats.
        """
        if value is None:
            CXDict.value.fset(
                self,
                {
                    "r": 0,
                    "g": 0,
                    "b": 0,
                    "a": 1.0,
                },
            )

        elif type(value) not in [CXRGBAColor, list, dict, str]:
            raise TypeError(
                "value must be a dict of {'r': int, 'g': int, 'b': int, "
                "'a': float} or a list of [r, g, b, a] or a string of "
                "rgba(int,int,int,float)"
            )

        else:
            if isinstance(value, str):
                if not CXRGBAColor.is_color_str(value):
                    raise ValueError(
                        "str RGBA values must be in the format"
                        " rgb(int,int,int,float)"
                    )
                else:
                    components = value.split(",")
                    r = components[0].strip().split("rgba(")[1]
                    g = components[1].strip()
                    b = components[2].strip()
                    a = components[3].strip().split(")")[0]

                    candidate = {
                        "r": int(r),
                        "g": int(g),
                        "b": int(b),
                        "a": float(a),
                    }

            elif isinstance(value, list):
                if not CXRGBAColor.is_color_list(value):
                    raise ValueError(
                        "list RGBA values must be in the format" " (int,int,int,float)"
                    )
                else:
                    candidate = {
                        "r": value[0],
                        "g": value[1],
                        "b": value[2],
                        "a": value[3],
                    }

            elif isinstance(value, dict):
                if not CXRGBAColor.is_color_dict(value):
                    raise ValueError(
                        "RGBA dict must have three int values for keys"
                        " 'r', 'g', 'b' and a float value for key 'a'"
                    )
                else:
                    candidate = dict(value)

            else:
                candidate = deepcopy(value.value)

            CXDict.value.fset(self, candidate)

    def render(self) -> Any:
        """
        Renders the value in a form suitable for use in preparing Javascript.
        Typically, this will be the native `value`.

        :returns: `Any`
            A version of the `value` most appropriate for use in prepating the
            Javascript rendering.
        """
        r = self.value["r"]
        g = self.value["g"]
        b = self.value["b"]
        a = self.value["a"]
        return {self.label: f"rgba({r},{g},{b},{a})"}

    def __init__(self, label: str, value: Union["CXRGBAColor", dict, list, str]):
        """
        Initializes a new CXRGBAColor object using the RGBA value from an
         existing `CXRGBAColor` object, or a `dict`, `list`, or `string`
         following the Javascript `rgba()` format.
        :param value: `Union['CXRGBAColor', dict, list, str]`
            The value to be accepted.  See the `is_color_*()` methods for
            acceptable formats.
        """
        super().__init__(label, {"r": 0, "g": 0, "b": 0, "a": 1})
        self.value = value

    def __str__(self) -> str:
        """
        *str* function.  Converts the object into a JSON string.
        """
        return json.dumps(self.render())

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXRGBAColor object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        r = self.value["r"]
        g = self.value["g"]
        b = self.value["b"]
        a = self.value["a"]
        return (
            f"CXRGBAColor("
            f" label={json.dumps(self.label)},"
            f" value='rgba({r},{g},{b},{a})'"
            f")"
        )


@deprecated(action="ignore")
class CXRGBColor(CXDict):
    """
    A `CXConfig` object that manages `str` Javascript rgb() values.
    """

    @staticmethod
    def is_color_str(value: str):
        """
        A static method that evaluates a given string to see if it represents a
        Javascript rgb() statement.
        :param value: `str`
            A string to evaluate.  A valid Javascript value has the form
            `rgb(r, g, b)` where RGB values are `int` from 0-255.
        :returns: `bool`
            True if the string represents a Javascript rgb() statement.
        """
        if isinstance(value, str):
            if not value.startswith("rgb"):
                return False

            try:
                components = value.split(",")
                if len(components) != 3:
                    return False

                r = components[0].strip().split("rgb(")[1]
                g = components[1].strip()
                b = components[2].strip().split(")")[0]

                for x in [int(r), int(g), int(b)]:
                    if (x < 0) or (x > 255):
                        return False

                return True

            except:
                return False

        else:
            return False

    @staticmethod
    def is_color_list(value: list):
        """
        A static method that evaluates a given list to see if it represents a
        Javascript rgb() statement.
        :param value: `list`
            A list to evaluate.  A valid Javascript value has the form
            `rgb(r, g, b)` where RGB values are `int` from 0-255.
        :returns: `bool`
            True if the list represents a Javascript rgb() statement.
        """
        if isinstance(value, list):
            list_len = len(value)
            all_rgb_elements_int = all(isinstance(x, int) for x in value[:2])
            if (not list_len == 3) or (not all_rgb_elements_int):
                return False

            for x in value[:3]:
                if (x < 0) or (x > 255):
                    return False

            return True

        else:
            return False

    @staticmethod
    def is_color_dict(value: dict):
        """
        A static method that evaluates a given dict to see if it represents a
        Javascript rgb() statement.
        :param value: `dict`
            A dict to evaluate.  A valid Javascript value has the form
            `rgba(r, g, b)` where RGB values are `int` from 0-255.  For the dict
             to be valid its keys must be lower case r, g, and b characters.
        :returns: `bool`
            True if the dict represents a Javascript rgb() statement.
        """
        if isinstance(value, dict):
            list_len = len(value.keys())
            all_rgb_elements = all(
                x
                in [
                    "r",
                    "g",
                    "b",
                ]
                for x in value.keys()
            )
            all_rgb_elements_int = all(
                isinstance(value[x], int) for x in value.keys() if x in ["r", "g", "b"]
            )
            keys_rgb = all([j in ["r", "g", "b"] for j in [k for k in value.keys()]])
            if (
                (not list_len == 3)
                or (not all_rgb_elements)
                or (not all_rgb_elements_int)
                or (not keys_rgb)
            ):
                return False

            for key in value.keys():
                if key in ["r", "g", "b"]:
                    if (value[key] < 0) or (value[key] > 255):
                        return False

            return True

        else:
            return False

    @CXDict.value.setter
    def value(self, value: Union["CXRGBColor", dict, list, str]) -> None:
        """
        Sets the RGB value from an existing `CXRGBColor` object, or a `dict`,
        `list`, or `string` following the Javascript `rgb()` format.
        :param value: `Union['CXRGBColor', dict, list, str]`
            The value to be accepted.  See the `is_color_*()` methods for
            acceptable formats.
        """
        if value is None:
            CXDict.value.fset(
                self,
                {
                    "r": 0,
                    "g": 0,
                    "b": 0,
                },
            )

        elif type(value) not in [CXRGBColor, list, dict, str]:
            raise TypeError(
                "value must be a dict of {'r': int, 'g': int, 'b': int}"
                " or a list of [r, g, b] or a string of "
                "rgb(int,int,int)"
            )

        else:
            if isinstance(value, str):
                if not CXRGBColor.is_color_str(value):
                    raise ValueError(
                        "str RGB values must be in the format" " rgb(int,int,int)"
                    )
                else:
                    components = value.split(",")
                    r = components[0].strip().split("rgb(")[1]
                    g = components[1].strip()
                    b = components[2].strip().split(")")[0]

                    candidate = {
                        "r": int(r),
                        "g": int(g),
                        "b": int(b),
                    }

            elif isinstance(value, list):
                if not CXRGBColor.is_color_list(value):
                    raise ValueError(
                        "list RGB values must be in the format" " (int,int,int)"
                    )
                else:
                    candidate = {
                        "r": value[0],
                        "g": value[1],
                        "b": value[2],
                    }

            elif isinstance(value, dict):
                if not CXRGBColor.is_color_dict(value):
                    raise ValueError(
                        "RGB dict must have three int values for keys" " 'r', 'g', 'b'"
                    )
                else:
                    candidate = dict(value)

            else:
                candidate = deepcopy(value.value)

            CXDict.value.fset(self, candidate)

    def render(self) -> dict:
        """
        Renders the value in a form suitable for use in preparing Javascript.
        Typically, this will be the native `value`.

        :returns: `Any`
            A version of the `value` most appropriate for use in prepating the
            Javascript rendering.
        """
        r = self.value["r"]
        g = self.value["g"]
        b = self.value["b"]
        return {self.label: f"rgb({r},{g},{b})"}

    def __init__(self, label: str, value: Union["CXRGBColor", dict, list, str]):
        """
        Initializes a new CXRGBColor object using the RGB value from an
         existing `CXRGBColor` object, or a `dict`, `list`, or `string`
         following the Javascript `rgb()` format.
        :param value: `Union['CXRGBColor', dict, list, str]`
            The value to be accepted.  See the `is_color_*()` methods for
            acceptable formats.
        """
        super().__init__(label, {"r": 0, "g": 0, "b": 0})
        self.value = value

    def __str__(self) -> str:
        """
        *str* function.  Converts the object into a JSON string.
        """
        return json.dumps(self.render())

    def __repr__(self) -> str:
        """
        *repr* function.  Converts the CXRGBColor object into a pickle string
        that can be used with `eval` to establish a copy of the object.
        :returns: `str` An evaluatable representation of the object.
        """
        r = self.value["r"]
        g = self.value["g"]
        b = self.value["b"]
        return (
            f"CXRGBColor("
            f" label={json.dumps(self.label)},"
            f" value='rgb({r},{g},{b})'"
            f")"
        )


class CXGraphTypeOptions(Enum):
    """
    A set of known chart types permitted for use with CanvasXpress objects.  If
    a chart not yet identified in this list is required then use a `CXString`
    object with the label `graphType` and the value set to the name of the
    chart to be used.
    """

    Area = "Area"
    AreaLine = "AreaLine"
    Bar = "Bar"
    BarLine = "BarLine"
    Boxplot = "Boxplot"
    Bubble = "Bubble"
    Chord = "Chord"
    Circular = "Circular"
    Contour = "Contour"
    Correlation = "Correlation"
    Dashboard = "Dashboard"
    Density = "Density"
    Donnut = "Donnut"
    DotLine = "DotLine"
    Dotplot = "Dotplot"
    Facet = "Facet"
    Fish = "Fish"
    Gantt = "Gantt"
    Genome = "Genome"
    Heatmap = "Heatmap"
    Histogram = "Histogram"
    Kaplan_Meier = "Kaplan - Meier"
    Layout = "Layout"
    Line = "Line"
    Map = "Map"
    Meter = "Meter"
    Network = "Network"
    NonLinear_Fit = "NonLinear - Fit"
    Oncoprint = "Oncoprint"
    ParallelCoordinates = "ParallelCoordinates"
    Pie = "Pie"
    Radar = "Radar"
    Remote_Graphs = "Remote - Graphs"
    Ridge_Line = "Ridge - Line"
    SPLOM = "SPLOM"
    Sankey = "Sankey"
    Scatter2D = "Scatter2D"
    Scatter3D = "Scatter3D"
    ScatterBubble2D = "ScatterBubble2D"
    Stacked = "Stacked"
    StackedLine = "StackedLine"
    StackedPercent = "StackedPercent"
    StackedPercentLine = "StackedPercentLine"
    Sunburst = "Sunburst"
    TCGA = "TCGA"
    TagCloud = "TagCloud"
    Tree = "Tree"
    Treemap = "Treemap"
    Venn = "Venn"
    Violin = "Violin"


class CXGraphType(CXString):
    """
    A CXString that is aware of CanvasXpress types of graphs, such as 'Bar'.
    """

    CX_ATTRIBUTE = "graphType"

    @CXString.value.setter
    def value(self, value: Union[CXGraphTypeOptions, str]) -> None:
        """
        Sets the value using a known CanvasXpress option.
        """
        if value is None:
            raise ValueError("value cannot be None.")

        elif isinstance(value, CXGraphTypeOptions):
            CXString.value.fset(self, value.value)

        elif str(value) not in list(map(str, CXGraphTypeOptions.__members__)):
            raise ValueError("value must be a known type.")

        else:
            CXString.value.fset(self, str(value))

    def set_custom_value(self, value: str):
        """
        Permits a js value to be set, such as if a new option is recently
        made available that the Python framework is yet to be aware of.
        :param value: `str`
            The string value to set.
        """
        CXString.value.fset(self, value)

    def render(self) -> dict:
        """
        Renders the value in a form suitable for use in preparing Javascript.
        Typically, this will be the native `value`.

        :returns: `dict`
            A version of the `value` most appropriate for use in prepating the
            Javascript rendering.
        """
        return {self.label: self.value}

    def __init__(self, type: Union[CXGraphTypeOptions, str] = CXGraphTypeOptions.Bar):
        """
        Initializes a new CXGraphType object with a value corresponding to one
        of the values provided by `CXGraphTypeOptions`.
        """
        if isinstance(type, CXGraphTypeOptions):
            super().__init__(self.CX_ATTRIBUTE, type.value)

        else:
            super().__init__(self.CX_ATTRIBUTE, type)



from canvasxpress.canvas import CanvasXpress
from canvasxpress.render.popup import CXBrowserPopup

import pandas as pd

# Each file is a tab-delimited matrix with columns and an index, and we can
# use pandas to read each at their URL to create dataframes

y_data_url = "https://www.canvasxpress.org/data/cX-heatmapR-dat.txt"
y_data = pd.read_csv(y_data_url, sep="\t", index_col=0)

x_data_url = "https://www.canvasxpress.org/data/cX-heatmapR-smp.txt"
x_data = pd.read_csv(x_data_url, sep="\t", index_col=0)

z_data_url = "https://www.canvasxpress.org/data/cX-heatmapR-var.txt"
z_data = pd.read_csv(z_data_url, sep="\t", index_col=0)

# The dict structure is an easy way to shape and submit data to CanvasXpress
hm_data = {
    # Data
    "y": {
        # Samples, which can be considered the columns
        "smps": y_data.columns.to_list(),
        # Variables, which can be considered the index
        "vars": y_data.index.tolist(),
        # The data itself, split into rows per variable and ordered by samples
        "data": y_data.values.tolist(),
    },
    # Sample annotations
    "x": {col: x_data[col].to_list() for col in x_data.columns.to_list()},
    # Variable annotations
    "z": {col: z_data[col].to_list() for col in z_data.columns.to_list()},
}

if __name__ == "__main__":

    chart = CanvasXpress(
        # Data can be structured many ways, and here we use a dict
        data=hm_data,
        # Configuration options are key-value pairs of formatting instructions
        config={
            "graphType": "Heatmap",
            "title": "Overlays in Heatmap",
            "colorSmpDendrogramBy": "Treatment",
            "colorSpectrum": ["magenta", "blue", "black", "red", "gold"],
            "colorSpectrumZeroValue": 0,
            "heatmapIndicatorHeight": 50,
            "heatmapIndicatorHistogram": True,
            "heatmapIndicatorPosition": "topLeft",
            "heatmapIndicatorWidth": 60,
            "heatmapSmpSeparateBy": "Treatment",
            "samplesClustered": True,
            "smpOverlays": ["Treatment", "Site"],
            "variablesClustered": True,
        },
        # The effect depends on the renderer, and for browser popups these set
        # the canvas element width and height
        width=675,
        height=600,
    )

    browser = CXBrowserPopup(chart)
    browser.render()
